# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable problem - Design space
*******************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import ones, zeros

from .variables import (
    X_SHARED_NAME,
    check_consistency,
    get_coupling_name,
    get_x_local_name,
)

LOGGER = logging.getLogger(__name__)


class TMDesignSpace(object):
    """The design space for the scalable problem introduced by Tedford and Martins
    (2010) defines the lower and upper bounds of both local design parameters, shared
    design parameters and coupling variables, as well as default values.

    The lower bounds are all equal to 0, the upper bounds are all equal to 1 and default
    values are all equal to 0.5.
    """

    def __init__(
        self,
        n_shared=1,
        n_local=None,
        n_coupling=None,
        default_inputs=None,
        dtype="float64",
    ):
        """The construction of the design space requires the number of shared design
        parameters, the number of local design parameters per discipline and the number
        of coupling variables per discipline. The two latter arguments must be list of
        integers with the same length which corresponds to the number of strongly
        coupled disciplines. By default, the design space considers two disciplines.

        :param int n_shared: size of the shared design parameters.
            Default: 1.
        :param list(int) n_local: sizes of the local design parameters.
            If None, use [1, 1]. Default: None.
        :param list(int) n_coupling: sizes of the coupling parameters.
            If None, use [1, 1]. Default: None.
        :param dict default_inputs: default inputs.
            Default: None.
        :param dtype: numpy data type. Default: 'float64'.
        """
        # Set and check the dimensions of the problem
        n_local = n_local or [1, 1]
        n_coupling = n_coupling or [1, 1]
        check_consistency(n_shared, n_local, n_coupling)

        default_inputs = {} if not isinstance(default_inputs, dict) else default_inputs
        self.names = []
        self.sizes = {}
        self.lower_bounds = {}
        self.upper_bounds = {}
        self.default_values = {}

        # Set the local design parameters
        for index, size in enumerate(n_local):
            name = get_x_local_name(index)
            value = default_inputs.get(name, zeros(size) + 0.5)
            value.astype(dtype)
            self.names.append(name)
            self.sizes[name] = size
            self.lower_bounds[name] = zeros(size)
            self.upper_bounds[name] = ones(size)
            self.default_values[name] = value

        # Set the shared design parameters
        name = X_SHARED_NAME
        value = default_inputs.get(X_SHARED_NAME, zeros(n_shared) + 0.5)
        value.astype(dtype)
        self.names.append(name)
        self.sizes[name] = n_shared
        self.lower_bounds[name] = zeros(n_shared)
        self.upper_bounds[name] = ones(n_shared)
        self.default_values[name] = value

        # Set the coupling variables
        for index, size in enumerate(n_coupling):
            name = get_coupling_name(index)
            value = default_inputs.get(name, zeros(size) + 0.5)
            value.astype(dtype)
            self.names.append(name)
            self.sizes[name] = size
            self.lower_bounds[name] = zeros(size)
            self.upper_bounds[name] = ones(size)
            self.default_values[name] = value



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard, Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Functions, f(x), from disciplines execution
*******************************************
"""
from __future__ import division, unicode_literals

import logging
from multiprocessing import Value
from numbers import Number

from numpy import abs as np_abs
from numpy import (
    absolute,
    add,
    array,
    atleast_1d,
    atleast_2d,
    concatenate,
    empty,
    hstack,
    matmul,
    multiply,
    ndarray,
    ones_like,
    reshape,
    subtract,
    vstack,
    where,
    zeros,
    zeros_like,
)
from numpy.linalg import multi_dot, norm
from six import string_types

from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.derivatives_approx import ComplexStep, FirstOrderFD

LOGGER = logging.getLogger(__name__)


class MDOFunction(object):
    """A container object to represent a function in an MDO process."""

    TYPE_OBJ = "obj"
    TYPE_EQ = "eq"
    TYPE_INEQ = "ineq"
    TYPE_OBS = "obs"
    AVAILABLE_TYPES = [TYPE_OBJ, TYPE_EQ, TYPE_INEQ, TYPE_OBS]
    DICT_REPR_ATTR = ["name", "f_type", "expr", "args", "dim", "special_repr"]
    DEFAULT_ARGS_BASE = "x"
    INDEX_PREFIX = "!"
    COEFF_FORMAT_1D = "{:.2e}"  # ensure that coefficients strings have same length
    COEFF_FORMAT_ND = "{: .2e}"  # ensure that coefficients strings have same length
    # N.B. the space character ensures same length whatever the sign of the coefficient

    def __init__(
        self,
        func,
        name,
        f_type=None,
        jac=None,
        expr=None,
        args=None,
        dim=None,
        outvars=None,
        force_real=False,
        special_repr=None,
    ):
        """Initializes the function attributes.

        :param func: the pointer to the function to be actually called
        :type func: callable
        :param name: the name of the function as a string
        :type name: str
        :param f_type: the type of function among (obj, eq, ineq, obs)
        :type f_type: str, optional
        :param jac: the jacobian
        :type jac: callable, optional
        :param expr: the function expression
        :type expr: str, optional
        :param args: the function arguments
        :type args: list(str), optional
        :param dim: the output dimension
        :type dim: int, optional
        :param outvars: the array of variable names used as inputs
            of the function
        :type outvars: list(str), optional
        :param force_real: if True, cast the results to real value
        :type force_real: bool, optional
        """
        super(MDOFunction, self).__init__()
        self._n_calls = Value("i", 0)
        # Initialize attributes
        self._f_type = None
        self._func = NotImplementedCallable()
        self._jac = NotImplementedCallable()
        self._name = None
        self._args = None
        self._expr = None
        self._dim = None
        self._outvars = None
        # Use setters to check values
        self.func = func
        self.jac = jac
        self.name = name
        self.f_type = f_type
        self.expr = expr
        self.args = args
        self.dim = dim
        self.outvars = outvars
        self.last_eval = None
        self.force_real = force_real
        self.special_repr = special_repr

    @property
    def n_calls(self):
        """Returns the number of calls to execute() which triggered the _run()
        multiprocessing safe."""
        return self._n_calls.value

    @n_calls.setter
    def n_calls(self, value):
        """Sets the number of calls to execute() which triggered the _run()
        multiprocessing safe."""
        with self._n_calls.get_lock():
            self._n_calls.value = value

    @property
    def func(self):
        """Accessor to the func property."""
        return self.__counted_f

    def __counted_f(self, x_vect):
        """Calls the function in self[self.SCIPY_FUN_TAG]

        :param x_vect: the function argument
        """
        with self._n_calls.get_lock():
            self._n_calls.value += 1
        val = self._func(x_vect)
        self.last_eval = val
        return val

    @func.setter
    def func(self, f_pointer):
        """Sets the f pointer.

        :param f_pointer: the pointer to the function to be actually called
        """
        self._n_calls.value = 0
        self._func = f_pointer

    def __call__(self, x_vect):
        """Calls the function.

        :param x_vect: the function argument
        """
        val = self.evaluate(x_vect, self.force_real)
        return val

    def evaluate(self, x_vect, force_real=False):
        """Evaluate the function.

        :param x_vect: the function argument
        :type x_vect: ndarray
        :param force_real: if True, cast the results to real value
        :type force_real: bool, optional
        :return: the function value
        :rtype: Number or ndarray
        """
        val = self.__counted_f(x_vect)
        if force_real:
            val = val.real
        self.dim = atleast_1d(val).shape[0]
        return val

    @property
    def name(self):
        """Accessor to the name of the function."""
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of the function.

        :param name: the name of the functioon
        """
        if name is not None and not isinstance(name, string_types):
            raise TypeError(
                "MDOFunction name must be a string, got " + str(type(name)) + " instead"
            )
        self._name = name

    @property
    def f_type(self):
        """Accessor to the function type."""
        return self._f_type

    @f_type.setter
    def f_type(self, f_type):
        """Sets the function type.

        :param f_type: the type of function among MDOFunction.AVAILABLE_TYPES
        """
        if f_type is not None and f_type not in self.AVAILABLE_TYPES:
            raise ValueError(
                "MDOFunction type must be among "
                + str(self.AVAILABLE_TYPES)
                + " , got "
                + str(f_type)
                + " instead"
            )
        self._f_type = f_type

    @property
    def jac(self):
        """Accessor to the jacobian."""
        return self._jac

    @jac.setter
    def jac(self, jac):
        """Sets the function jacobian.

        :param jac: pointer to a jacobian function
        """
        if jac is not None:
            if not callable(jac):
                raise TypeError("Jacobian function must be callable")
        self._jac = jac

    @property
    def args(self):
        """Accessor to the arguments list."""
        return self._args

    @args.setter
    def args(self, args):
        """Sets the function arguments.

        :param args: arguments list
        """
        if args is not None:
            if isinstance(args, ndarray):
                self._args = args.tolist()
            else:
                self._args = list(args)
        else:
            self._args = None

    @property
    def expr(self):
        """Accessor to the expression."""
        return self._expr

    @expr.setter
    def expr(self, expr):
        """Sets the function expression.

        :param expr: arguments list
        """
        if expr is not None and not isinstance(expr, string_types):
            raise TypeError(
                "Expression must be a string. "
                + " Got "
                + str(type(expr))
                + " instead."
            )
        self._expr = expr

    @property
    def dim(self):
        """Accessor to the dimension."""
        return self._dim

    @dim.setter
    def dim(self, dim):
        """Sets the function dimension.

        :param dim: int
        """
        if dim is not None and not int(dim) == dim:
            raise TypeError(
                "Dimension must be a integer. " + " Got " + str(type(int)) + " instead."
            )
        self._dim = dim

    @property
    def outvars(self):
        """Accessor to the array of input variable names used by the function."""
        return self._outvars

    @outvars.setter
    def outvars(self, outvars):
        """Sets the array of output variable names used by the function.

        :param outvars: array of variable names
        """
        if outvars is not None:
            if isinstance(outvars, ndarray):
                self._outvars = outvars.tolist()
            else:
                self._outvars = list(outvars)
        else:
            self._outvars = None

    def is_constraint(self):
        """Returns True if self.f_type is eq or ineq.

        :return: True if and only if the function is a contraint
        :rtype: bool
        """
        return self.f_type in [self.TYPE_EQ, self.TYPE_INEQ]

    def __repr__(self):
        """Self representation as a string."""
        return self.special_repr or self.default_repr

    @property
    def default_repr(self):
        str_repr = self.name
        if self.has_args():
            arguments = ", ".join(self.args)
            str_repr += "({})".format(arguments)

        if self.has_expr():
            str_repr += " = "
            expr = self.expr
            n_char = len(str_repr)
            # Remove empty lines with filter
            expr_spl = [_f for _f in expr.split("\n") if _f]
            str_repr += expr_spl[0]
            for repre in expr_spl[1:]:
                str_repr += "\n" + " " * n_char + repre
        return str_repr

    def has_jac(self):
        """Check if MDOFunction has a jacobian.

        :returns: True if self has a jacobian
        :rtype: bool
        """
        return self.jac is not None and not isinstance(
            self._jac, NotImplementedCallable
        )

    def has_dim(self):
        """Check if MDOFunction has a dimension.

        :returns: True if self has a dimension
        :rtype: bool
        """
        return self.dim is not None

    def has_outvars(self):
        """Check if MDOFunction has an array of input variables.

        :returns: True if self has a dimension
        :rtype: bool
        """
        return self.outvars is not None

    def has_expr(self):
        """Check if MDOFunction has an expression.

        :returns: True if self has an expression
        :rtype: bool
        """
        return self.expr is not None

    def has_args(self):
        """Check if MDOFunction has an args.

        :returns: True if self has an args
        :rtype: bool
        """
        return self.args is not None

    def has_f_type(self):
        """Check if MDOFunction has an type.

        :returns: True if self has a type
        :rtype: bool
        """
        return self.f_type is not None

    def apply_operator(self, other_f, operator, operator_repr):
        """Defines addition/substraction  for MDOFunction Supports automatic
        differentiation if other_f and self have a Jacobian.

        :param other_f: param operator:
        :param operator_repr: the representation as a string
        :param operator: the operator as a function pointer
        """
        if isinstance(other_f, MDOFunction):
            return self.__apply_operator_tofunction(other_f, operator, operator_repr)

        if isinstance(other_f, Number):
            return self.__apply_operator_to_number(other_f, operator, operator_repr)

        raise TypeError(
            "Unuspported + operand for MDOFunction and " + str(type(other_f))
        )

    def __apply_operator_tofunction(self, other_f, operator, operator_repr):
        """Defines addition/substraction  for MDOFunction Supports automatic
        differentiation if other_f and self have a Jacobian.

        :param other_f: param operator:
        :param operator_repr: the representation as a string
        :param operator: the operator as a function pointer
        """
        assert isinstance(other_f, MDOFunction)

        def add_f_pt(x_vect):
            """Addition function.

            :param x_vect: design variables
            :returns: self(x_vect) added to other_f(x_vect)
            """
            selfval = self(x_vect)
            otherval = other_f(x_vect)
            return operator(selfval, otherval)

        add_name = self.name + operator_repr + other_f.name
        self_oper_f = MDOFunction(
            add_f_pt,
            add_name,
            args=self.args,
            f_type=self.f_type,
            dim=self.dim,
            outvars=self.outvars,
        )

        if self.has_jac() and other_f.has_jac():

            def add_jac(x_vect):
                """Jacobian of addition function.

                :param x_vect: design variables
                :returns: Jac self(x_vect) added to Jac other_f(x_vect)
                """
                self_jac = self._jac(x_vect)
                other_jac = other_f.jac(x_vect)
                return operator(self_jac, other_jac)

            self_oper_f.jac = add_jac

        if self.has_expr() and other_f.has_expr():
            f_expr = self.expr + operator_repr + other_f.expr
            self_oper_f.expr = f_expr

        if self.has_args() and other_f.has_args():
            args = sorted(list(set(self.args + other_f.args)))
            self_oper_f.args = args

        if self.has_f_type():
            self_oper_f.f_type = self.f_type
        elif other_f.has_f_type():
            self_oper_f.f_type = other_f.f_type

        return self_oper_f

    def __apply_operator_to_number(self, other_val, operator, operator_repr):
        """Defines addition/substraction  for MDOFunction Supports automatic
        differentiation if other_f and self have a Jacobian.

        :param other_val: param operator: the value to add/subs
        :param operator_repr: the representation as a string
        :param operator: the operator as a function pointer
        """
        assert isinstance(other_val, Number)

        def add_f_pt(x_vect):
            """Addition function.

            :param x_vect: design variables
            :returns: self(x_vect) added to other_f(x_vect)
            """
            selfval = self(x_vect)
            return operator(selfval, other_val)

        add_name = self.name + operator_repr + str(other_val)
        self_oper_f = MDOFunction(
            add_f_pt,
            add_name,
            args=self.args,
            f_type=self.f_type,
            dim=self.dim,
            outvars=self.outvars,
        )

        if self.has_jac():
            self_oper_f.jac = self.jac

        if self.has_expr():
            self_oper_f.expr = self.expr + operator_repr + str(other_val)

        if self.has_args():
            self_oper_f.args = self.args

        if self.has_f_type():
            self_oper_f.f_type = self.f_type

        return self_oper_f

    def __add__(self, other_f):
        """Defines addition for MDOFunction Supports automatic differentiation if
        other_f and self have a jacobian."""
        return self.apply_operator(other_f, operator=add, operator_repr="+")

    def __sub__(self, other_f):
        """Defines subtraction for MDOFunction Supports automatic differentiation if
        other_f and self have a jacobian."""
        return self.apply_operator(other_f, operator=subtract, operator_repr="-")

    def __neg__(self):
        """Defines minus operator for MDOFunction (-f) Supports automatic
        differentiation if other_f and self have a Jacobian."""

        def min_pt(x_vect):
            """Negative self function.

            :param x_vect: design variables
            :returns: self(x_vect)
            """
            return -self(x_vect)

        min_name = "-" + self.name
        min_self = MDOFunction(
            min_pt,
            min_name,
            args=self.args,
            f_type=self.f_type,
            dim=self.dim,
            outvars=self.outvars,
        )

        if self.has_jac():

            def min_jac(x_vect):
                """Jacobian of negative self.

                :param x_vect: design variables
                :returns: Jac of -self(x_vect)
                """
                return -self.jac(x_vect)  # pylint: disable=E1102

            min_self.jac = min_jac

        if self.has_expr():
            min_self.expr = self.expr
            min_self.expr = min_self.expr.replace("+", "++")
            min_self.expr = min_self.expr.replace("-", "+")
            min_self.expr = min_self.expr.replace("++", "-")
            min_self.expr = "-" + min_self.expr

        return min_self

    def __mul__(self, other):
        """Defines multiplication for MDOFunction Supports automatic linearization if
        other_f and self have a jacobian."""
        is_number = isinstance(other, Number)
        is_func = isinstance(other, MDOFunction)
        if not is_number and not is_func:
            raise TypeError(
                "Unuspported * operand for MDOFunction and " + str(type(other))
            )

        if is_func:

            def mul_f_pt(x_vect):
                """Multiplication function.

                :param x_vect: design variables
                :returns: self(x_vect) added to other_f(x_vect)
                """
                return self(x_vect) * other(x_vect)

            out_name = self.name + "*" + other.name

        else:  # Func is a number

            def mul_f_pt(x_vect):
                """Multiplication function.

                :param x_vect: design variables
                :returns: self(x_vect) added to other_f(x_vect)
                """
                return self(x_vect) * other

            out_name = self.name + "*" + str(other)

        self_oper_f = MDOFunction(
            mul_f_pt,
            out_name,
            args=self.args,
            f_type=self.f_type,
            dim=self.dim,
            outvars=self.outvars,
        )
        if is_func:
            if self.has_jac() and other.has_jac():

                def mult_jac(x_vect):
                    """Jacobian of multiplication function.

                    :param x_vect: design variables
                    :returns: Jac self(x_vect) added to Jac other_f(x_vect)
                    """
                    self_f = self(x_vect)
                    other_f = other(x_vect)
                    self_jac = self._jac(x_vect)
                    other_jac = other.jac(x_vect)
                    return self_jac * other_f + other_jac * self_f

                self_oper_f.jac = mult_jac
        elif self.has_jac():

            def mult_jac(x_vect):
                """Jacobian of multiplication function.

                :param x_vect: design variables
                :returns: Jac self(x_vect) added to Jac other_f(x_vect)
                """
                return self._jac(x_vect) * other

            self_oper_f.jac = mult_jac

        if self.has_expr():
            if is_func and other.has_expr():
                f_expr = "(" + self.expr + ")*(" + other.expr + ")"
                self_oper_f.expr = f_expr
            if is_number:
                f_expr = "(" + self.expr + ")*" + str(other)
                self_oper_f.expr = f_expr

        if is_func and self.has_args() and other.has_args():
            args = sorted(list(set(self.args + other.args)))
            self_oper_f.args = args

        if self.has_f_type():
            self_oper_f.f_type = self.f_type
        elif is_func and other.has_f_type():
            self_oper_f.f_type = other.f_type

        return self_oper_f

    def offset(self, value):
        """Adds an offset value to the function.

        :param value: the offset value
        :type value: Number or ndarray
        :returns: the offset function as an MDOFunction
        :rtype: MDOFunction
        """

        def wrapped_function(x_vect):
            """Wrapped provided function in order to give to optimizer.

            :param x_vect: design variables
            :returns: evaluation of function at design variables
                     plus the offset
            """
            return self(x_vect) + value

        expr = self.expr
        if expr is not None:
            if value < 0:
                expr += " - " + str(-value)
            else:
                expr += " + " + str(abs(value))
        else:
            offset_is_negative = value < 0
            if hasattr(offset_is_negative, "__len__"):
                offset_is_negative = offset_is_negative.all()
            if offset_is_negative:
                expr = self.name + " - " + str(-value)
            else:
                expr = self.name + " + " + str(abs(value))
        offset_func = MDOFunction(
            wrapped_function,
            name=self.name,
            f_type=self.f_type,
            expr=expr,
            args=self.args,
            dim=self.dim,
            jac=self.jac,
            outvars=self.outvars,
        )

        return offset_func
        # Nothing to do for Jacobian since it is unchanged.

    def restrict(
        self,
        frozen_indexes,
        frozen_values,
        input_dim,
        name=None,
        f_type=None,
        expr=None,
        args=None,
    ):
        r"""Return a restriction of the function

        :math:`\newcommand{\frozeninds}{I}\newcommand{\xfrozen}{\hat{x}}\newcommand{
        \frestr}{\hat{f}}`
        For a subset :math:`\approxinds` of the variables indexes of a function
        :math:`f` to remain frozen at values :math:`\xfrozen_{i \in \frozeninds}` the
        restriction of :math:`f` is given by

        .. math::
            \frestr:
            x_{i \not\in \approxinds}
            \longmapsto
            f(\xref_{i \in \approxinds}, x_{i \not\in \approxinds}).

        :param frozen_indexes: indexes of the variables that will be frozen
        :type frozen_indexes: ndarray
        :param frozen_values: values of the variables that will be frozen
        :type frozen_values: ndarray
        :param input_dim: dimension of the function input (before restriction)
        :type input_dim: int
        :param name: name of the restriction
        :type name: str, optional
        :param f_type: type of the restriction
        :type f_type: str, optional
        :param expr: the expression of the restriction
        :type expr: str, optional
        :param args: arguments names of the restriction
        :type args: list(str), optional
        :return: restriction of the function
        :rtype: MDOFunction
        """
        # Check the shapes of the passed arrays
        if frozen_indexes.shape != frozen_values.shape:
            raise ValueError("Arrays of frozen indexes and values must have same shape")

        active_indexes = array([i for i in range(input_dim) if i not in frozen_indexes])

        def extend_subvect(x_subvect):
            """Return the extension of a restriction input with the frozen values.

            :param x_subvect: input of the restriction
            :type x_subvect: ndarray
            :return: extended input
            :rtype: ndarray
            """
            x_vect = empty(input_dim)
            x_vect[active_indexes] = x_subvect
            x_vect[frozen_indexes] = frozen_values
            return x_vect

        def func(x_subvect):
            """Return the value of the restriction.

            :param x_subvect: input of the restriction
            :type x_subvect: ndarray
            :return: value of the restriction
            :rtype: float
            """
            x_vect = extend_subvect(x_subvect)
            value = self.__call__(x_vect)
            return value

        # Build the name of the restriction
        if name is None and args is not None:
            name = self.name + "_wrt_" + "_".join(args)
        elif name is None:
            name = self.name + "_restriction"

        restriction = MDOFunction(
            func,
            name,
            f_type,
            expr=expr,
            args=args,
            dim=self.dim,
            outvars=self.outvars,
            force_real=self.force_real,
        )

        if self.has_jac():

            def jac(x_subvect):
                """Return the Jacobian matrix of the restriction.

                :param x_subvect: input of the restriction
                :type x_subvect: ndarray
                :return: Jacobian matrix of the restriction
                :rtype: ndarray
                """
                x_vect = extend_subvect(x_subvect)
                full_jac = self.jac(x_vect)
                if len(full_jac.shape) == 1:
                    sub_jac = full_jac[active_indexes]
                else:
                    sub_jac = full_jac[:, active_indexes]
                return sub_jac

            restriction.jac = jac

        return restriction

    def linear_approximation(self, x_vect, name=None, f_type=None, args=None):
        r"""Return the first-order Taylor polynomial of the function at a given point

        :math:`\newcommand{\xref}{\hat{x}}\newcommand{\dim}{d}`
        The first-order Taylor polynomial of a (possibly vector-valued) function
        :math:`f` at a point :math:`\xref` is defined as

        .. math::
            \newcommand{\partialder}{\frac{\partial f}{\partial x_i}(\xref)}
            f(x)
            \approx
            f(\xref) + \sum_{i = 1}^{\dim} \partialder \, (x_i - \xref_i).

        :param x_vect: point defining the Taylor polynomial
        :type x_vect: ndarray
        :param name: name of the linear approximation
        :type name: str, optional
        :param f_type: the function type of the linear approximation
        :type f_type: str, optional
        :param args: names for each scalar variable, or a name base
        :type args: list(str), optional
        :returns: first-order Taylor polynomial of the function at the given point
        :rtype: MDOLinearFunction
        """
        # Check that the function Jacobian is available
        if not self.has_jac():
            raise AttributeError(
                "Function Jacobian unavailable for linear approximation"
            )

        # Build the first-order Taylor polynomial
        coefficients = self.jac(x_vect)
        func_val = self.__call__(x_vect)
        if isinstance(func_val, ndarray):
            # Make sure the function value is at most 1-dimensional
            func_val = func_val.flatten()
        value_at_zero = func_val - matmul(coefficients, x_vect)
        linear_approx_suffix = "_linearized"
        name = self.name + linear_approx_suffix if name is None else name
        args = args if args else self.args
        linear_approx = MDOLinearFunction(
            coefficients, name, f_type, args, value_at_zero
        )

        return linear_approx

    def convex_linear_approx(self, x_vect, approx_indexes=None, sign_threshold=1e-9):
        r"""Return the convex linearization of the function at a given point

        :math:`\newcommand{\xref}{\hat{x}}\newcommand{\dim}{d}`
        The convex linearization of a function :math:`f` at a point :math:`\xref`
        is defined as

        .. math::
            \newcommand{\partialder}{\frac{\partial f}{\partial x_i}(\xref)}
            f(x)
            \approx
            f(\xref)
            +
            \sum_{\substack{i = 1 \\ \partialder > 0}}^{\dim}
            \partialder \, (x_i - \xref_i)
            -
            \sum_{\substack{i = 1 \\ \partialder < 0}}^{\dim}
            \partialder \, \xref_i^2 \, \left(\frac{1}{x_i} - \frac{1}{\xref_i}\right).

        :math:`\newcommand{\approxinds}{I}`
        Optionally, one may require the convex linearization of :math:`f` with
        respect to a subset
        :math:`x_{i \in \approxinds} \subset \{x_1, \dots, x_{\dim}\}`
        of its variables rather than all of them:

        .. math::
            f(x)
            =
            f(x_{i \in \approxinds}, x_{i \not\in \approxinds})
            \approx
            f(\xref_{i \in \approxinds}, x_{i \not\in \approxinds})
            +
            \sum_{\substack{i \in \approxinds \\ \partialder > 0}}
            \partialder \, (x_i - \xref_i)
            -
            \sum_{\substack{i \in \approxinds \\ \partialder < 0}}
            \partialder
            \, \xref_i^2 \, \left(\frac{1}{x_i} - \frac{1}{\xref_i}\right).

        :param x_vect: point defining the convex linearization
        :type x_vect: ndarray
        :param approx_indexes: array of booleans specifying w.r.t. which variables
            the function should be approximated (by default, all of them)
        :type approx_indexes: ndarray, optional
        :param sign_threshold: threshold for the sign of the derivatives
        :type sign_threshold: float, optional
        :returns: convex linearization of the function at the given point
        :rtype: MDOFunction
        """
        # Check the approximation indexes
        if approx_indexes is None:
            approx_indexes = ones_like(x_vect, dtype=bool)
        elif approx_indexes.shape != x_vect.shape or approx_indexes.dtype != "bool":
            raise ValueError(
                "The approximation array must an array of booleans with "
                "the same shape as the function argument"
            )

        # Get the function Jacobian matrix
        if not self.has_jac():
            raise AttributeError(
                "Function Jacobian unavailable for convex linearization"
            )
        jac = atleast_2d(self.jac(x_vect))

        # Build the coefficients matrices
        coeffs = jac[:, approx_indexes]
        direct_coeffs = where(coeffs > sign_threshold, coeffs, 0.0)
        recipr_coeffs = multiply(
            -where(-coeffs > sign_threshold, coeffs, 0.0), x_vect[approx_indexes] ** 2
        )

        def get_steps(x_new):
            """Return the steps on the direct and reciprocal variables.

            :param x_new: argument of the convex linearization
            :type x_new: ndarray
            :returns: step on the direct variables, step on the reciprocal variables
            :rtype: ndarray, ndarray
            """
            step = x_new[approx_indexes] - x_vect[approx_indexes]
            inv_step = zeros_like(step)
            nonzero_indexes = (absolute(step) > sign_threshold).nonzero()
            inv_step[nonzero_indexes] = 1.0 / step[nonzero_indexes]
            return step, inv_step

        def convex_lin_func(x_new):
            """Return the value of the function convex linearization.

            :param x_new: argument of the convex linearization
            :type x_new: ndarray
            :returns: value of the function convex linearization
            :rtype: ndarray
            """
            merged_vect = where(approx_indexes, x_vect, x_new)
            step, inv_step = get_steps(x_new)
            value = (
                self.__call__(merged_vect)
                + matmul(direct_coeffs, step)
                + matmul(recipr_coeffs, inv_step)
            )
            if self._dim == 1:
                return value[0]
            return value

        def convex_lin_jac(x_new):
            """Return the Jacobian of the function convex linearization.

            :param x_new: argument of the convex linearization
            :type x_new: ndarray
            :returns: Jacobian of the function convex linearization
            :rtype: ndarray
            """
            merged_vect = where(approx_indexes, x_vect, x_new)
            value = atleast_2d(self.jac(merged_vect))
            _, inv_step = get_steps(x_new)
            value[:, approx_indexes] = direct_coeffs + multiply(
                recipr_coeffs, -(inv_step ** 2)
            )
            if self._dim == 1:
                value = value[0, :]
            return value

        # Build the convex linearization of the function
        convex_lin_approx = MDOFunction(
            convex_lin_func,
            self.name + "_convex_lin",
            self.f_type,
            convex_lin_jac,
            args=None,
            dim=self.dim,
            outvars=self.outvars,
            force_real=self.force_real,
        )

        return convex_lin_approx

    def quadratic_approx(self, x_vect, hessian_approx, args=None):
        r"""Return a quadratic appproximation of the (scalar-valued) function at a
        given point

        :math:`\newcommand{\xref}{\hat{x}}\newcommand{\dim}{d}\newcommand{
        \hessapprox}{\hat{H}}`
        For a given approximation :math:`\hessapprox` of the Hessian matrix of a
        function :math:`f` at a point :math:`\xref`, the quadratic approximation of
        :math:`f` is defined as

        .. math::
            \newcommand{\partialder}{\frac{\partial f}{\partial x_i}(\xref)}
            f(x)
            \approx
            f(\xref)
            + \sum_{i = 1}^{\dim} \partialder \, (x_i - \xref_i)
            + \frac{1}{2} \sum_{i = 1}^{\dim} \sum_{j = 1}^{\dim}
            \hessapprox_{ij} (x_i - \xref_i) (x_j - \xref_j).

        :param x_vect: point defining the Taylor polynomial
        :type x_vect: ndarray
        :param hessian_approx: approximation of the Hessian matrix at x_vect
        :type hessian_approx: ndarray
        :param args: names for each scalar variable, or a name base
        :type args: list(str), optional
        :returns: second-order Taylor polynomial of the function at the given point
        :rtype: MDOQuadraticFunction
        """
        # Build the second-order coefficients
        if (
            not isinstance(hessian_approx, ndarray)
            or hessian_approx.ndim != 2
            or hessian_approx.shape[0] != hessian_approx.shape[1]
        ):
            raise ValueError("Hessian approximation must be a square ndarray")
        if hessian_approx.shape[1] != x_vect.size:
            raise ValueError(
                "Hessian approximation and vector must have same dimension"
            )
        quad_coeffs = 0.5 * hessian_approx

        # Build the first-order coefficients
        if not self.has_jac():
            raise AttributeError("Jacobian unavailable")
        gradient = self.jac(x_vect)
        hess_dot_vect = matmul(hessian_approx, x_vect)
        linear_coeffs = gradient - hess_dot_vect

        # Buid the zero-order coefficient
        zero_coeff = matmul(0.5 * hess_dot_vect - gradient, x_vect) + self.__call__(
            x_vect
        )

        # Build the second-order Taylor polynomial
        quadratic_suffix = "_quadratized"
        name = self.name + quadratic_suffix
        args = args if args else self.args
        quad_approx = MDOQuadraticFunction(
            quad_coeffs=quad_coeffs,
            linear_coeffs=linear_coeffs,
            value_at_zero=zero_coeff,
            name=name,
            args=args,
        )

        return quad_approx

    def check_grad(self, x_vect, method="FirstOrderFD", step=1e-6, error_max=1e-8):
        """Checks the gradient of self.

        :param x_vect: the vector at which the function is checked
        :type x_vect: ndarray
        :param method: FirstOrderFD or ComplexStep
            (Default value = "FirstOrderFD")
        :type method: str, optional
        :param step: the step for approximation
            (Default value = 1e-6)
        :type step: float, optional
        :param error_max: Default value = 1e-8)
        :type error_max: float, optional
        """

        def rel_err(a_vect, b_vect):
            """Compute relative error.

            :param a_vect: param b_vect:
            :param b_vect:
            """
            n_b = norm(b_vect)
            if n_b > error_max:
                return norm(a_vect - b_vect) / norm(b_vect)
            return norm(a_vect - b_vect)

        def filt_0(arr, floor_value=1e-6):
            """Filtering of error.

            :param arr:
            :param floor_value:  (Default value = 1e-6)
            :param floor_value:  (Default value = 1e-6)
            """
            return where(np_abs(arr) < floor_value, 0.0, arr)

        if method == "FirstOrderFD":
            apprx = FirstOrderFD(self, step)
        elif method == "ComplexStep":
            apprx = ComplexStep(self, step)
        else:
            raise ValueError(
                "Unknwon approximation method "
                + str(method)
                + ", use FirstOrderFD or ComplexStep."
            )
        apprx_grad = apprx.f_gradient(x_vect).real
        anal_grad = self._jac(x_vect).real
        if not apprx_grad.shape == anal_grad.shape:
            shape_app1d = len(apprx_grad.shape) == 1 or apprx_grad.shape[0] == 1
            shape_ex1d = len(anal_grad.shape) == 1 or anal_grad.shape[0] == 1
            shape_1d = shape_app1d and shape_ex1d
            flatten_diff = anal_grad.flatten().shape != apprx_grad.flatten().shape
            if not shape_1d or (shape_1d and flatten_diff):
                raise ValueError(
                    "Inconsistent function jacobian shape ! "
                    + "Got :"
                    + str(anal_grad.shape)
                    + " while expected :"
                    + str(apprx_grad.shape)
                )
        rel_error = rel_err(anal_grad, apprx_grad)
        succeed = rel_error < error_max
        if not succeed:
            LOGGER.error("Function jacobian is wrong %s", str(self))
            LOGGER.error("Error =\n%s", str(filt_0(anal_grad - apprx_grad)))
            LOGGER.error("Analytic jacobian=\n%s", str(filt_0(anal_grad)))
            LOGGER.error("Approximate step gradient=\n%s", str(filt_0(apprx_grad)))
            raise ValueError("Function jacobian is wrong " + str(self))

    def get_data_dict_repr(self):
        """Returns a dict representation of self for serialization Pointers to functions
        are removed.

        :returns: a dict with attributes names as keys
        :rtype: dict
        """
        repr_dict = {}
        for attr_name in self.DICT_REPR_ATTR:
            attr = getattr(self, attr_name)
            if attr is not None:
                repr_dict[attr_name] = attr
        return repr_dict

    @staticmethod
    def init_from_dict_repr(**kwargs):
        """Initalizes a new Function from a data dict typically used for
        deserialization.

        :param kwargs: key value pairs from DICT_REPR_ATTR
        """
        allowed = MDOFunction.DICT_REPR_ATTR
        for key in kwargs:
            if key not in allowed:
                raise ValueError(
                    "Cannot initialize MDOFunction "
                    + "attribute: "
                    + str(key)
                    + ", allowed ones are: "
                    + ", ".join(allowed)
                )
        return MDOFunction(func=None, **kwargs)

    def set_pt_from_database(
        self, database, design_space, normalize=False, jac=True, x_tolerance=1e-10
    ):
        """self.__call__(x) returns f(x) if x is in the database and self.name in the
        database keys Idem for jac.

        :param database: the database to read
        :type database: Database
        :param design_space: the design space used for normalization
        :type design_space: DesignSpace
        :param normalize: if True, x_n is unnormalized before call
        :type normalize: bool, optional
        :param jac: if True, a jacobian pointer is also generated
        :type jac: bool, optional
        :param x_tolerance: tolerance on the distance between inputs
        :type x_tolerance: float, optional
        """
        name = self.name

        def read_in_db(x_n, fname):
            """Reads fname in the database for the x_n entry.

            :param x_n: the x to evaluate fname
            :returns : f(x_n) if present in the database
            """
            if normalize:
                x_db = design_space.unnormalize_vect(x_n)
            else:
                x_db = x_n
            val = database.get_f_of_x(fname, x_db, x_tolerance)
            if val is None:
                msg = "Function " + str(fname)
                msg += " evaluation relies only on the database, and"
                msg += str(fname) + "( x ) is not in the database "
                msg += " for x=" + str(x_db)
                raise ValueError(msg)
            return val

        def f_from_db(x_n):
            """Computes f(x_n) from the database.

            :param x_n: the x to evaluate f
            """
            return read_in_db(x_n, name)

        def j_from_db(x_n):
            """Computes df(x_n)/d x_n from the database.

            :param x_n: the x to evaluate the jacobian
            """
            return read_in_db(x_n, "@" + name)

        self.func = f_from_db
        if jac:
            self.jac = j_from_db

    @staticmethod
    def generate_args(input_dim, args=None):
        """Generate the strings for a function arguments.

        :param input_dim: number of scalar input arguments
        :type input_dim: int
        :param args: the initial function arguments strings
        :type args: list(str), optional
        :returns: the arguments strings
        :rtype: list(str)
        """
        if args and len(args) == input_dim:
            # Keep the passed list of strings
            new_args = args
        elif args and len(args) == 1:
            # Generate the arguments strings based on a the unique passed string
            new_args = MDOFunction._generate_args(args[0], input_dim)
        else:
            # Generate the arguments strings based on the default string
            args_base = "x"
            new_args = MDOFunction._generate_args(args_base, input_dim)

        return new_args

    @staticmethod
    def _generate_args(args_base, input_dim):
        """Generate the arguments strings based on a string.

        :param args_base: base string for the arguments strings
        :type args_base: str
        :param input_dim: number of scalar input arguments
        :type input_dim: int
        :returns: arguments strings
        :rtype: list(str)
        """
        index_prefix = MDOLinearFunction.INDEX_PREFIX
        return [args_base + index_prefix + str(i) for i in range(input_dim)]

    @staticmethod
    def concatenate(
        functions,
        name,
        f_type=None,
    ):
        """Concatenate functions.

        :param functions: functions to be concatenated
        :type functions: list(MDOFunction)
        :param name: name of the concatenation
        :type name: str
        :param f_type: type of the concatenation function
        :type f_type: str, optional
        :return: concatenation of the functions
        :rtype: MDOFunction
        """

        def concat_func(x_vect):
            """Concatenate functions values.

            :param x_vect: common input of the functions
            :type x_vect: ndarray
            :return: concatenation of the functions values
            :rtype: ndarray
            """
            return concatenate([atleast_1d(func(x_vect)) for func in functions])

        def concat_jac(x_vect):
            """Concatenate functions Jacobian matrices.

            :param x_vect: common input of the functions
            :type x_vect: ndarray
            :return: concatenation of the functions Jacobian matrices
            :rtype: ndarray
            """
            return vstack([atleast_2d(func.jac(x_vect)) for func in functions])

        dim = sum([func.dim for func in functions])
        outvars_list = [func.outvars for func in functions]
        if None in outvars_list:
            outvars = None
        else:
            outvars = [out_var for outvars in outvars_list for out_var in outvars]
        return MDOFunction(
            concat_func, name, f_type, concat_jac, dim=dim, outvars=outvars
        )


class MDOLinearFunction(MDOFunction):
    r"""Linear multivariate function defined by

    * a matrix :math:`A` of first-order coefficients
      :math:`(a_{ij})_{\substack{i = 1, \dots m \\ j = 1, \dots n}}`
    * and a vector :math:`b` of zero-order coefficients :math:`(b_i)_{i = 1, \dots m}`

    .. math::
        F(x)
        =
        Ax + b
        =
        \begin{bmatrix}
            a_{11} & \cdots & a_{1n} \\
            \vdots & \ddots & \vdots \\
            a_{m1} & \cdots & a_{mn}
        \end{bmatrix}
        \begin{bmatrix} x_1 \\ \vdots \\ x_n \end{bmatrix}
        +
        \begin{bmatrix} b_1 \\ \vdots \\ b_m \end{bmatrix}.

    """

    def __init__(self, coefficients, name, f_type=None, args=None, value_at_zero=0.0):
        """Initialize the function attributes.

        :param coefficients: coefficients of the linear function
        :type coefficients: ndarray or Number
        :param name: the name of the function
        :type name: str
        :param f_type: the type of function among (obj, eq, ineq)
        :type f_type: str, optional
        :param args: the function arguments
        :type args: list(str), optional
        :param value_at_zero: function value at zero
        :type value_at_zero: ndarray or Number, optional
        """
        # Format the passed coefficients and value at zero
        self.coefficients = coefficients
        output_dim, input_dim = self._coefficients.shape
        self.value_at_zero = value_at_zero

        # Generate the arguments strings
        new_args = MDOLinearFunction.generate_args(input_dim, args)

        # Generate the expression string
        if output_dim == 1:
            expr = self._generate_1d_expr(new_args)
        else:
            expr = self._generate_nd_expr(new_args)

        # Define the function as a linear combination with an offset
        def fun(x_vect):
            """Return the linear combination with an offset:

            sum_{i=1}^n a_i * x_i + b

            :param x_vect: design variables
            :type x_vect: ndarray
            """
            value = matmul(self._coefficients, x_vect) + self._value_at_zero
            if value.size == 1:
                value = value[0]
            return value

        # Jacobian is constant
        def jac(_):
            """Return the constant Jacobian, which is the matrix of coefficients. N.B.
            if the function is scalar a 1d-array is returned (the gradient).

            :returns: function Jacobian or function gradient
            :rtype: ndarray
            """
            if self._coefficients.shape[0] == 1:
                jac = self._coefficients[0, :]
            else:
                jac = self._coefficients
            return jac

        super(MDOLinearFunction, self).__init__(
            fun,
            name,
            f_type=f_type,
            jac=jac,
            expr=expr,
            args=new_args,
            dim=output_dim,
            outvars=[name],
        )

    @property
    def coefficients(self):
        """Get the function coefficients. N.B. shall return a 2-dimensional ndarray.

        :returns: function coefficients
        :rtype: ndarray
        """
        return self._coefficients

    @coefficients.setter
    def coefficients(self, coefficients):
        """Set the function coefficients. N.B. the coefficients shall be stored as a
        2-dimensional ndarray.

        :param coefficients: function coefficients
        :type coefficients: ndarray or Number
        """
        if isinstance(coefficients, Number):
            self._coefficients = atleast_2d(coefficients)
        elif isinstance(coefficients, ndarray) and len(coefficients.shape) == 2:
            self._coefficients = coefficients
        elif isinstance(coefficients, ndarray) and len(coefficients.shape) == 1:
            self._coefficients = coefficients.reshape([1, -1])
        else:
            raise ValueError(
                "Coefficients must be passed as a 2-dimensional or "
                "1-dimensional ndarray"
            )

    @property
    def value_at_zero(self):
        """Get the function value at zero. N.B. shall return a 1-dimensional ndarray.

        :returns: function value at zero
        :rtype: ndarray
        """
        return self._value_at_zero

    @value_at_zero.setter
    def value_at_zero(self, value_at_zero):
        """Set the function value at zero N.B. the value at zero shall be stored as a
        1-dimensional ndarray.

        :param value_at_zero: function value at zero
        :type value_at_zero: ndarray or Number
        """
        output_dim = self.coefficients.shape[0]  # N.B. the coefficients must be set
        if isinstance(value_at_zero, ndarray) and value_at_zero.size == output_dim:
            self._value_at_zero = value_at_zero.reshape(output_dim)
        elif isinstance(value_at_zero, Number):
            self._value_at_zero = array([value_at_zero] * output_dim)
        else:
            raise ValueError("Value at zero must be an ndarray or a number")

    def _generate_1d_expr(self, args):
        """Generate the literal expression of the linear function in scalar form.

        :param args: the function arguments
        :type args: list(str)
        :returns: function literal expression
        :rtype: str
        """
        expr = ""

        # Build the expression of the linear combination
        first_non_zero = -1
        for i, coeff in enumerate(self._coefficients[0, :]):
            if coeff != 0.0:
                if first_non_zero == -1:
                    first_non_zero = i
                # Add the monomial sign
                if i == first_non_zero and coeff < 0.0:
                    # The first nonzero coefficient is negative.
                    expr += "-"  # unary minus
                elif i != first_non_zero and coeff < 0.0:
                    expr += " - "
                elif i != first_non_zero and coeff > 0.0:
                    expr += " + "
                # Add the coefficient value
                if abs(coeff) != 1.0:
                    expr += MDOFunction.COEFF_FORMAT_1D.format(abs(coeff)) + "*"
                # Add argument string
                expr += args[i]

        # Add the offset expression
        value_at_zero = self._value_at_zero[0]
        if first_non_zero == -1:
            # Constant function
            expr += MDOFunction.COEFF_FORMAT_1D.format(value_at_zero)
        elif self._value_at_zero > 0.0:
            expr += " + " + MDOFunction.COEFF_FORMAT_1D.format(value_at_zero)
        elif self._value_at_zero < 0.0:
            expr += " - " + MDOFunction.COEFF_FORMAT_1D.format(-value_at_zero)

        return expr

    def _generate_nd_expr(self, args):
        """Generate the literal expression of the linear function in matrix form.

        :param args: the function arguments
        :type args: list(str)
        :returns: function literal expression
        :rtype: str
        """
        max_args_len = max([len(arg) for arg in args])
        out_dim, in_dim = self._coefficients.shape
        expr = ""
        for i in range(max(out_dim, in_dim)):
            if i > 0:
                expr += "\n"
            # matrix line
            if i < out_dim:
                line_coeffs_str = (
                    MDOFunction.COEFF_FORMAT_ND.format(coeff)
                    for coeff in self._coefficients[i, :]
                )
                expr += "[{}]".format(" ".join(line_coeffs_str))
            else:
                expr += " " + " ".join([" " * 3] * in_dim) + " "
            # vector line
            if i < in_dim:
                expr += "[{}]".format(args[i])
            else:
                expr += " " * (max_args_len + 2)
            # sign
            if i == 0:
                expr += " + "
            else:
                expr += "   "
            # value at zero
            if i < out_dim:
                expr += "[{}]".format(
                    MDOFunction.COEFF_FORMAT_ND.format(self._value_at_zero[i])
                )
        return expr

    def __neg__(self):
        """Define the minus operator for an MDOLinearFunction (-f).

        :returns: the opposite function
        :rtype: MDOLinearFunction
        """
        return MDOLinearFunction(
            -self._coefficients,
            "-" + self.name,
            self.f_type,
            self.args,
            -self._value_at_zero,
        )

    def offset(self, value):
        """Add an offset value to the linear function.

        :param value: the offset value
        :type value: Number or ndarray
        :returns: the offset function
        :rtype: MDOLinearFunction
        """
        return MDOLinearFunction(
            self._coefficients,
            self.name,
            self.f_type,
            self.args,
            self._value_at_zero + value,
        )

    def restrict(self, frozen_indexes, frozen_values):
        """Build a restriction of the linear function.

        :param frozen_indexes: indexes of the variables that will be frozen
        :type frozen_indexes: ndarray
        :param frozen_values: values of the variables that will be frozen
        :type frozen_values: ndarray
        :returns: the restriction of the linear function
        :rtype: MDOLinearFunction
        """
        # Check the shapes of the passed arrays
        if frozen_indexes.shape != frozen_values.shape:
            raise ValueError("Arrays of frozen indexes and values must have same shape")

        # Separate the frozen coefficients from the active ones
        frozen_coefficients = self.coefficients[:, frozen_indexes]
        active_indexes = array(
            [i for i in range(self.coefficients.shape[1]) if i not in frozen_indexes]
        )
        active_coefficients = self.coefficients[:, active_indexes]

        # Compute the restriction value at zero
        value_at_zero = matmul(frozen_coefficients, frozen_values) + self._value_at_zero

        # Build the function restriction
        restriction_suffix = "_restriction"
        name = self.name + restriction_suffix
        args = [self.args[i] for i in active_indexes]
        restriction = MDOLinearFunction(
            active_coefficients, name, args=args, value_at_zero=value_at_zero
        )

        return restriction


class MDOQuadraticFunction(MDOFunction):
    r"""Scalar-valued quadratic multivariate function defined by

    * a *square* matrix :math:`A` of second-order coefficients
      :math:`(a_{ij})_{\substack{i = 1, \dots n \\ j = 1, \dots n}}`
    * a vector :math:`b` of first-order coefficients :math:`(b_i)_{i = 1, \dots n}`
    * and a scalar zero-order coefficient :math:`c`

    .. math::
        f(x)
        =
        c
        +
        \sum_{i = 1}^n b_i \, x_i
        +
        \sum_{i = 1}^n \sum_{j = 1}^n a_{ij} \, x_i \, x_j.
    """

    def __init__(
        self,
        quad_coeffs,
        name,
        f_type=None,
        args=None,
        linear_coeffs=None,
        value_at_zero=None,
    ):
        """Initialize a quadratic function.

        :param quad_coeffs: second-order coefficients
        :type quad_coeffs: ndarray
        :param name: name of the function
        :type name: str
        :param f_type: type of the function
        :type f_type: str, optional
        :param args: arguments names of the function
        :type args: list(str), optional
        :param linear_coeffs: first-order coefficients
        :type linear_coeffs: ndarray, optional
        :param value_at_zero: zero-order coefficient
        :type value_at_zero: float, optional
        """
        self._input_dim = 0
        self._quad_coeffs = None
        self._linear_coeffs = None
        self.quad_coeffs = quad_coeffs  # sets the input dimension
        new_args = MDOFunction.generate_args(self._input_dim, args)

        # Build the first-order term
        if linear_coeffs is not None:
            linear_part = MDOLinearFunction(linear_coeffs, name + "_lin", args=new_args)
            self.linear_coeffs = linear_part.coefficients
        self._value_at_zero = value_at_zero

        def func(x_vect):
            """Compute the quadratic function value.

            :param x_vect: quadratic function argument
            :type x_vect: ndarray
            :returns: the quadratic function value
            :rtype: float
            """
            value = multi_dot((x_vect.T, self._quad_coeffs, x_vect))
            if self._linear_coeffs is not None:
                value += linear_part(x_vect)
            if self._value_at_zero is not None:
                value += self._value_at_zero
            return value

        def grad(x_vect):
            """Compute the quadratic function gradient.

            :param x_vect: quadratic function argument
            :type x_vect: ndarray
            :returns: the quadratic function gradient
            :rtype: ndarray
            """
            gradient = matmul(self._quad_coeffs + self._quad_coeffs.T, x_vect)
            if self._linear_coeffs is not None:
                gradient += linear_part.jac(x_vect)
            return gradient

        # Build the second-order term
        expr = self.build_expression(
            self._quad_coeffs, new_args, self._linear_coeffs, self._value_at_zero
        )
        super(MDOQuadraticFunction, self).__init__(
            func, name, f_type, grad, expr, args=new_args, dim=1
        )

    @property
    def quad_coeffs(self):
        """Get the function second-order coefficients N.B. shall return a 2-dimensional
        ndarray.

        :returns: function second-order coefficients
        :rtype: ndarray
        """
        return self._quad_coeffs

    @quad_coeffs.setter
    def quad_coeffs(self, coefficients):
        """Set the function second-order coefficients N.B. the coefficients shall be
        stored as a 2-dimensional ndarray.

        :param coefficients: function second-order coefficients
        :type coefficients: ndarray
        """
        # Check the second-order coefficients
        if (
            not isinstance(coefficients, ndarray)
            or len(coefficients.shape) != 2
            or coefficients.shape[0] != coefficients.shape[1]
        ):
            raise ValueError(
                "Quadratic coefficients must be passed as a 2-dimensional "
                "square ndarray"
            )
        self._quad_coeffs = coefficients
        self._input_dim = self._quad_coeffs.shape[0]

    @property
    def linear_coeffs(self):
        """Get the function first-order coefficients N.B. shall return a 1-dimensional
        ndarray.

        :returns: function first-order coefficients
        :rtype: ndarray
        """
        if self._linear_coeffs is None:
            return zeros(self._input_dim)
        return self._linear_coeffs

    @linear_coeffs.setter
    def linear_coeffs(self, coefficients):
        """Set the function first-order coefficients N.B. the coefficients shall be
        stored as a 1-dimensional ndarray.

        :param coefficients: function first-order coefficients
        :type coefficients: ndarray
        """
        if coefficients.size != self._input_dim:
            raise ValueError(
                "The number of first-order coefficients must be equal "
                "to the input dimension"
            )
        self._linear_coeffs = coefficients

    @staticmethod
    def build_expression(quad_coeffs, args, linear_coeffs=None, value_at_zero=None):
        """Return the expression of the quadratic function.

        :param quad_coeffs: second-order coefficients
        :type quad_coeffs: ndarray
        :param args: arguments names of the function
        :type args: list(str)
        :param linear_coeffs: first-order coefficients
        :type linear_coeffs: ndarray, optional
        :param value_at_zero: zero-order coefficient
        :type value_at_zero: float, optional
        """
        transpose_str = "'"
        expr = ""
        for index, line in enumerate(quad_coeffs.tolist()):
            arg = args[index]
            # Second-order expression
            line = quad_coeffs[index, :].tolist()
            expr += "[{}]".format(arg)
            expr += transpose_str if index == 0 else " "
            quad_coeffs_str = (MDOFunction.COEFF_FORMAT_ND.format(val) for val in line)
            expr += "[{}]".format(" ".join(quad_coeffs_str))
            expr += "[{}]".format(arg)
            # First-order expression
            if (
                linear_coeffs is not None
                and (linear_coeffs != zeros_like(linear_coeffs)).any()
            ):
                expr += " + " if index == 0 else "   "
                expr += "[{}]".format(
                    MDOFunction.COEFF_FORMAT_ND.format(linear_coeffs[0, index])
                )
                expr += transpose_str if index == 0 else " "
                expr += "[{}]".format(arg)
            # Zero-order expression
            if value_at_zero is not None and value_at_zero != 0.0 and index == 0:
                sign_str = "+" if value_at_zero > 0.0 else "-"
                expr += (" {} " + MDOFunction.COEFF_FORMAT_ND).format(
                    sign_str, abs(value_at_zero)
                )
            if index < quad_coeffs.shape[0] - 1:
                expr += "\n"
        return expr


class MDOFunctionGenerator(object):
    """The link between MDODisciplines and objective functions and constraints is made
    with MDOFunctionGenerator, which generates MDOFunctions from the disciplines.

    It uses closures to generate functions instances from a discipline execution.
    """

    def __init__(self, discipline):
        """Constructor.

        :param discipline: the discipline from which the generator
                           builds functions
        :type discipline: MDODiscipline
        """
        self.discipline = discipline

    def __make_function(self, input_names_list, output_names_list, default_inputs):
        """Makes a function from io and reference data Combines lambda function and a
        local def for function pointer issues Uses a closure.

        :param input_names_list: the dict keys of the input data
        :param output_names_list: the dict keys of the output data
        :param default_inputs: a default inputs dict to eventually overload
            the discipline's default inputs when evaluating the discipline
        :returns: The generated function
        """

        def func(x_vect):
            """A function which executes a discipline.

            :param x_vect: the input vector of the function
            :type x_vect: ndarray
            :returns: the selected outputs of the discipline
            :rtype: ndarray
            """
            for name in input_names_list:
                if name not in self.discipline.default_inputs:
                    raise ValueError(
                        "Discipline "
                        + str(self.discipline.name)
                        + " has no default_input named "
                        + str(name)
                        + ", while input is required"
                        + " by MDOFunction."
                    )
            defaults = self.discipline.default_inputs
            if default_inputs is not None:
                defaults.update(default_inputs)
            data = DataConversion.update_dict_from_array(
                defaults, input_names_list, x_vect
            )
            self.discipline.reset_statuses_for_run()
            computed_values = self.discipline.execute(data)
            values_array = DataConversion.dict_to_array(
                computed_values, output_names_list
            )
            if values_array.size == 1:  # Then the function is scalar
                return values_array[0]
            return values_array

        def func_jac(x_vect):
            """A function which linearizes a discipline.

            :param x_vect: the input vector of the function
            :type x_vect: ndarray
            :returns: the selected outputs of the discipline
            :rtype: ndarray
            """
            defaults = self.discipline.default_inputs
            n_dv = len(x_vect)
            data = DataConversion.update_dict_from_array(
                defaults, input_names_list, x_vect
            )
            self.discipline.linearize(data)

            grad_array = []
            for out_name in output_names_list:
                jac_loc = self.discipline.jac[out_name]
                grad_loc = DataConversion.dict_to_array(jac_loc, input_names_list)
                grad_output = hstack(grad_loc)
                if len(grad_output) > n_dv:
                    grad_output = reshape(grad_output, (grad_output.size // n_dv, n_dv))
                grad_array.append(grad_output)
            grad = vstack(grad_array).real
            if grad.shape[0] == 1:
                grad = grad.flatten()
                assert len(x_vect) == len(grad)
            return grad

        default_name = "_".join(output_names_list)
        mdo_func = MDOFunction(
            func,
            name=default_name,
            args=input_names_list,
            outvars=output_names_list,
            jac=func_jac,
        )
        return mdo_func

    def get_function(
        self,
        input_names_list,
        output_names_list,
        default_inputs=None,
        differentiable=True,
    ):
        """Builds a function from a discipline input and output lists.

        :param input_names_list: names of inputs of the disciplines
           to be inputs of the function
        :type input_names_list: list(str)
        :param output_names_list: names of outputs of the disciplines
            to be returned by the function
        :type output_names_list: list(str)
        :param default_inputs: a default inputs dict to eventually overload
            the discipline's default inputs when evaluating the discipline
        :type default_inputs: dict, optional
        :param differentiable: if True then inputs and outputs are added to the list
            of variables to be differentiated
        :type differentiable: bool, optional
        :returns: the function
        :rtype: MDOFunction
        """
        if isinstance(input_names_list, string_types):
            input_names_list = [input_names_list]

        if isinstance(output_names_list, string_types):
            output_names_list = [output_names_list]

        if input_names_list is None:
            input_names_list = self.discipline.get_input_data_names()
        if output_names_list is None:
            output_names_list = self.discipline.get_output_data_names()

        if not self.discipline.is_all_inputs_existing(input_names_list):
            raise ValueError(
                "Some elements of "
                + str(input_names_list)
                + " are not inputs of the discipline, "
                + str(self.discipline.name)
                + " available inputs are:"
                + str(self.discipline.get_input_data_names())
            )

        if not self.discipline.is_all_outputs_existing(output_names_list):
            raise ValueError(
                "Some elements of "
                + str(output_names_list)
                + " are not outputs of the discipline, "
                + str(self.discipline.name)
                + " available outputs are: "
                + ", ".join(self.discipline.get_output_data_names())
                + "."
            )

        # adds inputs and outputs to the list of variables to be
        # differentiated
        if differentiable:
            self.discipline.add_differentiated_inputs(input_names_list)
            self.discipline.add_differentiated_outputs(output_names_list)

        return self.__make_function(input_names_list, output_names_list, default_inputs)


class NotImplementedCallable(object):
    """A not implemented callable object."""

    def __call__(self, *args, **kwargs):
        raise NotImplementedError("Function is not implemented")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#         documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Post-processing for scalability study.

The :class:`.PostScalabilityStudy` class implements the way as the set of
:class:`.ScalabilityResult`-based result files
contained in the study directory are graphically post-processed. This class
provides several methods to easily change graphical properties, notably
the plot labels. It also makes it possible to define a cost function per
MDO formulation, converting the numbers of executions and linearizations
of the different disciplines required by a MDO process in an estimation
of the computational cost associated with what would be a scaled version
of the true problem.

.. warning::

   Comparing MDO formulations in terms of estimated true computational time
   rather than CPU time of the :class:`.ScalabilityStudy` is highly
   recommended.
   Indeed, time is often an obviousness criterion to distinguish between
   MDO formulations having the same performance in terms of distance to the
   optimum: look at our calculation budget and choose the best formulation
   that satisfies this budget, or even saves us time. Thus, it is important
   to carefully define these cost functions.
"""

from __future__ import division, unicode_literals

import logging

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from numpy import array, atleast_3d
from numpy import bool as np_bool
from numpy import median, poly1d, polyfit

from gemseo.problems.scalable.data_driven.study.result import ScalabilityResult
from gemseo.utils.py23_compat import Path, string_types
from gemseo.utils.string_tools import MultiLineString

LOGGER = logging.getLogger(__name__)

CURRENT_DIRECTORY = Path.cwd()
PLOT_EXTENSION = "png"
RESULTS_DIRECTORY = Path("results")
POST_DIRECTORY = Path("visualization")
POSTSTUDY_DIRECTORY = POST_DIRECTORY / "scalability_study"


class PostScalabilityStudy(object):

    """The PostScalabilityStudy class aims to post-process a list of scalability results
    stored in a directory."""

    NOMENCLATURE = {
        "exec_time": "Execution time (s)",
        "original_exec_time": "Pseudo-original execution time",
        "n_calls": "Number of discipline evaluations",
        "n_calls_linearize": "Number of gradient evaluations",
        "status": "Optimizatin status",
        "is_feasible": "Feasibility of the solution (0 or 1)",
        "scaling_strategy": "Scaling strategy index",
        "total_calls": "Total number of evaluations",
    }

    def __init__(self, study_directory):
        """Constructor.

        :param str study_directory: directory of the scalability study."
        """
        msg = MultiLineString()
        msg.add("Post-process for scalability study")
        msg.indent()
        msg.add("Working directory: {}", study_directory)
        LOGGER.info("%s", msg)
        self.study_directory = Path(study_directory)
        self.scalability_results = self.__load_results()
        self.n_results = len(self.scalability_results)
        self.descriptions = self.NOMENCLATURE
        self.cost_function = {}
        self.unit_cost = None
        for result in self.scalability_results:
            result.total_calls = sum(result.n_calls.values())
            result.total_calls += sum(result.n_calls_linearize.values())

    def set_cost_function(self, formulation, cost):
        """Set cost function for each formulation.

        :param str formulation: name of the formulation.
        :param function cost: cost function
        """
        self.cost_function[formulation] = cost

    def set_cost_unit(self, cost_unit):
        """Set the measurement unit for cost evaluation.

        :param str cost_unit: cost unit, e.g. 'h', 'min', ...
        """
        self.unit_cost = cost_unit
        description = self.descriptions["original_exec_time"].split(" (")[0]
        description = "{} ({})".format(description, cost_unit)
        self.descriptions["original_exec_time"] = description

    def labelize_exec_time(self, description):
        """Change the description of execution time, used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("exec_time", description)

    def labelize_original_exec_time(self, description):
        """Change the description of original execution time, used in plotting methods
        notably.

        :param str description: description.
        """
        self._update_descriptions("original_exec_time", description)
        if self.unit_cost is not None:
            self.set_cost_unit(self.unit_cost)

    def labelize_n_calls(self, description):
        """Change the description of number of calls, used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("n_calls", description)

    def labelize_n_calls_linearize(self, description):
        """Change the description of number of calls for linearization, used in plotting
        methods notably.

        :param str description: description.
        """
        self._update_descriptions("n_calls_linearize", description)

    def labelize_status(self, description):
        """Change the description of status, used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("status", description)

    def labelize_is_feasible(self, description):
        """Change the description of feasibility, used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("is_feasible", description)

    def labelize_scaling_strategy(self, description):
        """Change the description of scaling strategy, used in plotting methods notably.

        :param str description: description.
        """
        self._update_descriptions("scaling_strategy", description)

    def _update_descriptions(self, keyword, description):
        """Update the description initialized with the NOMENCLATURE class attribute.

        :param str keyword: keyword of the considered object.
        :param str description: new description
        """
        if not self.descriptions.get(keyword):
            keywords = ", ".join(list(self.descriptions.keys()))
            raise ValueError(
                "The keyword {} is not in the list: {}".format(keyword, keywords)
            )
        if not isinstance(description, string_types):
            raise TypeError(
                'The argument "description" must be '
                "of type string, "
                "not of type {}".format(description.__class__.__name__)
            )
        self.descriptions[keyword] = description

    def __load_results(self):
        """Load results from the results directory of the study path."""
        if not self.study_directory.is_dir():
            raise ValueError(
                'Directory "{}" does not exist.'.format(self.study_directory)
            )
        directory = self.study_directory / RESULTS_DIRECTORY
        if not directory.is_dir():
            raise ValueError('Directory "{}" does not exist.'.format(directory))
        filenames = [
            filename.name for filename in directory.iterdir() if filename.is_file()
        ]
        results = []
        for filename in filenames:
            name = filename.split(".")[0]
            id_scaling = int(name.split("_")[-2])
            id_sample = int(name.split("_")[-1])
            result = ScalabilityResult(name, id_scaling, id_sample)
            result.load(self.study_directory)
            results.append(result)
        if not results:
            raise ValueError("Directory {} is empty.".format(directory))
        return results

    def plot(
        self,
        legend_loc="upper left",
        xticks=None,
        xticks_labels=None,
        xmargin=0.0,
        **options
    ):
        """Plot the results using different methods according to the presence or absence
        of replicate values.

        :param str legend_loc: legend localization
        :param list(float) xticks: list of xticks (default: None)
        :param list(str) xticks_labels: list of xticks labels (default: None)
        :param float xmargin: margin on left and right sides of the x-axis
        :param options: options for the specialized plot methods
        """
        msg = MultiLineString()
        msg.add("Execute post-processing")
        msg.indent()
        self.__has_scaling_dimension(xticks)
        self.__has_scaling_dimension(xticks_labels)
        there_are_replicates = len(self.get_replicates(True)) > 1
        if self.cost_function:
            self._estimate_original_time()
        if there_are_replicates:
            msg.add("Type: replicate")
            self._plot_boxes(legend_loc, xticks, xticks_labels, xmargin, **options)
        else:
            msg.add("Type: standard")
            self._plot_lines(legend_loc, xticks, xticks_labels, xmargin, **options)
        LOGGER.info("%s", msg)

    def __has_scaling_dimension(self, value):
        """Assert if a value has the scaling dimension.

        :param ndarray value: value.
        """
        if value is not None and hasattr(value, "__len__"):
            assert len(value) == len(self.get_scaling_strategies(True))

    def _plot_lines(
        self, legend_loc="upper left", xticks=None, xticks_labels=None, xmargin=0.0
    ):
        """Deterministic plot.

        :param str legend_loc: legend localization
        :param list(float) xticks: list of xticks (default: None)
        :param list(str) xticks_labels: list of xticks labels (default: None)
        :param float xmargin: margin on left and right sides of the x-axis
        """
        colors = ["blue", "red", "green"]
        handles = [
            Line2D([0], [0], color=colors[idx], lw=4, label=optim_strategy)
            for idx, optim_strategy in enumerate(self.optimization_strategies)
        ]
        for index, strategy in enumerate(self.optimization_strategies):
            color = colors[index]
            scales, criteria = self.__get_scales_and_criteria(strategy)
            for name, value in criteria.items():
                self.__draw(name, value, xticks, xticks_labels, scales, color, xmargin)
        for criterion in criteria:
            plt.figure(criterion)
            plt.legend(handles=handles, loc=legend_loc, frameon=False, framealpha=0.5)
            fname = Path(criterion).with_suffix("." + PLOT_EXTENSION)
            fpath = self.study_directory / POSTSTUDY_DIRECTORY / fname
            plt.savefig(str(fpath))

    def __draw(self, name, value, xticks, labels, scales, color, xmargin):
        """Create plot for specific criterion when r=1 replicate.

        :param str name: criterion name.
        :param str value: criterion value.
        :param list(float) xticks: list of xticks (default: None)
        :param list(str) labels: list of xticks labels (default: None)
        :param list(str) scales: scales
        :param str color: line color.
        :param float xmargin: margin on left and right sides of the x-axis
        """
        plt.figure(name)
        value = atleast_3d(value.T).T
        n_lines = value.shape[0]
        for id_line in range(n_lines):
            yvalues = value[id_line, :, 0].flatten()
            xvalues = scales
            linestyle = "-" if n_lines == 1 else (0, (id_line + 1, n_lines))
            if xticks is not None:
                xvalues = xticks
            else:
                xticks = xvalues
            plt.plot(xvalues, yvalues, "o", color=color)
            coef = polyfit(xvalues, yvalues, 1)
            poly1d_fn = poly1d(coef)
            plt.plot(xvalues, poly1d_fn(xvalues), linestyle=linestyle, color=color)
            if labels is None:
                labels = xticks
            else:
                labels = xvalues
            plt.xticks(xticks, labels)
            plt.xlabel(self.descriptions["scaling_strategy"])
            plt.ylabel(self.descriptions[name])
            plt.xlim(xvalues[0] - xmargin, xvalues[-1] + xmargin)
            plt.grid(True, "both")

    @property
    def n_samples(self):
        """Number of samples."""
        return len(self.get_replicates(True))

    def _plot_boxes(
        self,
        legend_loc="upper left",
        xticks=None,
        xticks_labels=None,
        xmargin=0.0,
        minbox=2,
        notch=False,
        widths=0.25,
        whis=1.5,
    ):
        """Probabilistic plot.

        :param str legend_loc: legend localization
        :param list(float) xticks: list of xticks (default: None)
        :param list(str) xticks_labels: list of xticks labels (default: None)
        :param float xmargin: margin on left and right sides of the x-axis
        :param int minbox: minimal number of values for boxplot (default: 2).
        :param bool notch: if True, will produced a notched boxplot.
        :param float whis:  the reach of the whiskers to the beyond
            the first and third quartiles (default: 1.5).
        """
        if not hasattr(widths, "__len__"):
            widths = [widths] * len(self.get_scaling_strategies(True))
        else:
            self.__has_scaling_dimension(widths)

        colors = ["blue", "red", "green"]
        for index, strategy in enumerate(self.optimization_strategies):
            color = colors[index]
            scales, criteria = self.__get_scales_and_criteria(strategy)
            for name, value in criteria.items():
                self.__drawb(
                    name,
                    index,
                    value,
                    xticks,
                    xticks_labels,
                    scales,
                    widths,
                    color,
                    notch,
                    whis,
                    xmargin,
                )
        handles = [
            Line2D([0], [0], color=colors[idx], lw=4, label=strategy)
            for idx, strategy in enumerate(self.optimization_strategies)
        ]
        for criterion in criteria:
            plt.figure(criterion)
            plt.legend(
                handles=handles,
                loc=legend_loc,
                frameon=False,
                framealpha=0.5,
            )
            fname = Path(criterion).with_suffix("." + PLOT_EXTENSION)
            fpath = self.study_directory / POSTSTUDY_DIRECTORY / fname
            indentation = MultiLineString.INDENTATION
            LOGGER.info("%sSave %s plot in %s", indentation, criterion, fpath)
            plt.savefig(str(fpath))

    def __drawb(
        self,
        name,
        index,
        value,
        xticks,
        labels,
        scales,
        widths,
        color,
        notch,
        whis,
        xmargin,
    ):
        """Create plot for specific criterion when r=1 replicate.

        :param str name: criterion name.
        :param int index: strategy index.
        :param str value: criterion value.
        :param list(float) xticks: list of xticks (default: None)
        :param list(str) labels: list of xticks labels (default: None)
        :param list(str) scales: scales
        :param str color: line color.
        :param bool notch: if True, will produced a notched boxplot.
        :param float whis:  the reach of the whiskers to the beyond
            the first and third quartiles (default: 1.5).
        :param float xmargin: margin on left and right sides of the x-axis
        """
        plt.figure(name)
        xvalues = scales
        if xticks is not None:
            xvalues = xticks
        n_strategies = len(self.optimization_strategies)
        tmp = []
        for idx, xtick in enumerate(xvalues):
            tmp.append(float(index) / n_strategies)
            tmp[-1] *= widths[idx] * 3
            tmp[-1] += xtick
        self.__draw_boxplot(value, tmp, color, color, notch, widths, whis)
        zvalues = atleast_3d(value.T).T[0]
        xval_offset = [
            xtick + float(index) / n_strategies * widths[idx] * 3
            for idx, xtick in enumerate(xvalues)
        ]
        plt.plot(xval_offset, median(zvalues, 1), "--", color=color)
        if labels is None:
            labels = xvalues
        plt.xticks(xvalues, labels)
        plt.xlim(xvalues[0] - xmargin, xvalues[-1] + xmargin)
        plt.xlabel(self.descriptions["scaling_strategy"])
        plt.ylabel(self.descriptions[name])
        plt.grid(True, "both")

    @staticmethod
    def __draw_boxplot(data, xticks, edge_color, fill_color, notch, widths, whis):
        """Draw boxplot from a dataset.

        :param array data: dataset array of dimension 2 or 3
        :param list(float) xticks: values of xticks
        :param str edge_color: edge color
        :param str fill_color: fill color
        :param bool notch: if True, will produced a notched boxplot.
        :param list(float) widths: widths of boxplots
        :param float whis:  the reach of the whiskers to the beyond
            the first and third quartiles
        """
        if len(data.shape) == 3:
            data = data[0, :, :]

        if data.dtype == np_bool:
            # To prevent error when arrays are substracted with recent numpy
            data = data.astype(int)

        boxplot = plt.boxplot(
            data.T,
            patch_artist=True,
            positions=xticks,
            showfliers=False,
            whiskerprops={"linestyle": "-"},
            notch=notch,
            widths=widths,
            whis=whis,
        )

        for element in ["boxes", "whiskers", "fliers", "means", "caps"]:
            plt.setp(boxplot[element], color=edge_color)
        plt.setp(boxplot["medians"], color="white")

        for patch in boxplot["boxes"]:
            patch.set(facecolor=fill_color)

        return boxplot

    def __get_scales_and_criteria(self, optim_strategy):
        """Get values of criteria and corresponding scaling levels for a given
        optimization strategy and all replicates.

        :param str optim_strategy: name of the optimization strategy.
        :return: scaling levels, criteria values
        :rtype: list(int), dict(array)
        """
        exec_time = []
        original_exec_time = []
        n_calls = []
        n_calls_linearize = []
        status = []
        is_feasible = []
        scaling = []
        total_calls = []
        for replicate in self.get_replicates(True):
            xcoord, ycoord = self.__get_replicate_values(optim_strategy, replicate)
            exec_time.append(ycoord["exec_time"])
            total_calls.append(ycoord["total_calls"])
            if self.cost_function:
                original_exec_time.append(ycoord["original_exec_time"])
            tmp = [
                list(ycoord["n_calls"][idx].values())
                for idx in range(len(ycoord["n_calls"]))
            ]
            n_calls.append(tmp)
            tmp = [
                list(ycoord["n_calls_linearize"][idx].values())
                for idx in range(len(ycoord["n_calls_linearize"]))
            ]
            n_calls_linearize.append(tmp)
            status.append(ycoord["status"])
            is_feasible.append(ycoord["is_feasible"])
            scaling.append(ycoord["scaling"])
        scaling_levels = xcoord
        values = {"exec_time": array(exec_time).T}  # (n_scal, n_rep)
        if self.cost_function:
            values["original_exec_time"] = array(original_exec_time).T  # (n_s, n_r)
        values["n_calls"] = array(n_calls).T  # (n_d, n_s, n_rep)
        values["n_calls_linearize"] = array(n_calls_linearize).T  # (n_d, n_s, n_r)
        values["total_calls"] = array(total_calls).T  # (n_scal, n_rep)
        values["is_feasible"] = array(is_feasible).T  # (n_scal, n_rep)
        return scaling_levels, values

    def __get_replicate_values(self, optim_strategy, replicate):
        """Get values of criteria and corresponding scaling levels for a given
        optimization strategy and a given replicate.

        :param str optim_strategy: optimization strategy.
        :param int replicate: replicate index.
        :return: scaling levels, criteria values
        :rtype: list(int), dict(array-like)
        """
        are_replicate = [value == replicate for value in self.get_replicates()]
        are_optim_strategy = [
            value == optim_strategy for value in self.get_optimization_strategies()
        ]
        are_ok = [
            is_rep and is_oo for is_rep, is_oo in zip(are_replicate, are_optim_strategy)
        ]
        indices = [index for index, is_ok in enumerate(are_ok) if is_ok]
        scaling_levels = self.get_scaling_strategies()
        scaling_levels = [scaling_levels[index] for index in indices]
        tmp = sorted(list(range(len(scaling_levels))), key=lambda k: scaling_levels[k])
        indices = [indices[index] for index in tmp]
        scaling_levels = [scaling_levels[index] for index in tmp]
        results = [self.scalability_results[index] for index in indices]
        exec_time = [result.exec_time for result in results]
        total_calls = [result.total_calls for result in results]
        original_exec_time = [result.original_exec_time for result in results]
        n_calls = [result.n_calls for result in results]
        n_calls_linearize = [result.n_calls_linearize for result in results]
        status = [result.status for result in results]
        is_feasible = [result.is_feasible for result in results]
        scaling = [result.scaling for result in results]
        values = {
            "exec_time": exec_time,
            "total_calls": total_calls,
            "n_calls": n_calls,
            "n_calls_linearize": n_calls_linearize,
            "status": status,
            "is_feasible": is_feasible,
            "scaling": scaling,
        }
        if self.cost_function:
            values["original_exec_time"] = original_exec_time
        return scaling_levels, values

    @property
    def names(self):
        """Get the names of the scalability results."""
        return [value.name for value in self.scalability_results]

    def get_optimization_strategies(self, unique=False):
        """Get the names of the optimization strategies.

        :param bool unique: return either unique values if True
            or one value per scalability result if False (default: False).
        :return: names of the optimization stategies.
        :rtype: list(str)
        """
        os_list = ["_".join(name.split("_")[0:-2]) for name in self.names]
        if unique:
            os_list = sorted(set(os_list))
        return os_list

    @property
    def optimization_strategies(self):
        return self.get_optimization_strategies(True)

    def get_scaling_strategies(self, unique=False):
        """Get the identificants of the scaling strategies.

        :param bool unique: return either unique values if True
            or one value per scalability result if False (default: False).
        :return: identifiants of scaling strategies
        :rtype: list(int)
        """
        ss_list = [int(name.split("_")[-2]) for name in self.names]
        if unique:
            ss_list = sorted(set(ss_list))
        return ss_list

    def get_replicates(self, unique=False):
        """Get the replicate identifiants.

        :param bool unique: return either unique values if True
            or one value per scalability result if False (default: False).
        :return: identifiants of replicates.
        :rtype: list(int)
        """
        rep = [int(name.split("_")[-1]) for name in self.names]
        if unique:
            rep = sorted(set(rep))
        return rep

    def _estimate_original_time(self):
        """Estimate the original execution time from the number of calls and
        linearizations of the different disciplines and top-level disciplines and from
        the cost functions provided by the user.

        :return: original time
        :rtype: float
        """
        for scalability_result in self.scalability_results:
            n_c = scalability_result.n_calls
            n_cl = scalability_result.n_calls_linearize
            n_tl_c = scalability_result.n_calls_top_level
            n_tl_cl = scalability_result.n_calls_linearize_top_level
            varsizes = scalability_result.new_varsizes
            formulation = scalability_result.formulation
            if formulation not in self.cost_function:
                raise ValueError(
                    "The cost function of {} must be defined "
                    "in order to compute "
                    "the estimated original time.".format(formulation)
                )
            result = self.cost_function[formulation](
                varsizes, n_c, n_cl, n_tl_c, n_tl_cl
            )
            scalability_result.original_exec_time = result



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class to create a probability distribution from the SciPy library.

The :class:`.SPDistribution` class is a concrete class
inheriting from :class:`.Distribution` which is an abstract one.
SP stands for `scipy <https://docs.scipy.org/doc/scipy/reference/
tutorial/stats.html>`_ which is the library it relies on.

The :class:`.SPDistribution` of a given uncertain variable is built
from mandatory arguments:

- a variable name,
- a distribution name recognized by SciPy,
- a set of parameters provided as a dictionary
  of keyword arguments named as the arguments of the scipy constructor
  of this distribution.

.. warning::

    The distribution parameters must be provided according to the signature
    of the scipy classes. `Access the scipy documentation
    <https://docs.scipy.org/doc/scipy/reference/stats.html>`_.

The constructor has also optional arguments:

- a variable dimension (default: 1),
- a standard representation of these parameters
  (default: use :code:`parameters`).
"""

from __future__ import division, unicode_literals

import logging
from typing import Callable, Iterable, List, Mapping, Optional, Union

import scipy
import scipy.stats as sp_stats
from numpy import array, ndarray, vstack

from gemseo.uncertainty.distributions.distribution import (
    Distribution,
    ParametersType,
    StandardParametersType,
)
from gemseo.uncertainty.distributions.scipy.composed import SPComposedDistribution

LOGGER = logging.getLogger(__name__)

SP_WEBSITE = "https://docs.scipy.org/doc/scipy/reference/stats.html"


class SPDistribution(Distribution):
    """SciPy probability distribution.

    Create a probability distribution for an uncertain variable
    from its dimension and distribution names and properties.

    .. seealso::

        :class:`.SPExponentialDistribution`
        :class:`.SPNormalDistribution`
        :class:`.SPTriangularDistribution`
        :class:`.SPUniformDistribution`

    Example:
        >>> from gemseo.uncertainty.distributions.scipy.distribution import (
        ...    SPDistribution
        ... )
        >>> distribution = SPDistribution('x', 'expon', {'loc': 3, 'scale': 1/2.})
        >>> print(distribution)
        expon(loc=3, scale=0.5)
    """

    _COMPOSED_DISTRIBUTION = SPComposedDistribution

    def __init__(
        self,
        variable,  # type: str
        interfaced_distribution,  # type: str
        parameters,  # type: ParametersType
        dimension=1,  # type: int
        standard_parameters=None,  # type: Optional[StandardParametersType]
    ):  # noqa: D205,D212,D415
        # type: (...) -> None
        """
        Args:
            variable: The name of the random variable.
            interfaced_distribution: The name of the probability distribution,
                typically the name of a class wrapped from an external library,
                such as 'Normal' for OpenTURNS or 'norm' for SciPy.
            parameters: The parameters of the probability distribution.
            dimension: The dimension of the random variable.
            standard_parameters (dict, optional): The standard representation
                of the parameters of the probability distribution.
        """
        super(SPDistribution, self).__init__(
            variable,
            interfaced_distribution,
            parameters,
            dimension,
            standard_parameters,
        )
        self.marginals = self.__create_distributions(
            self.distribution_name, self.parameters
        )
        math_support = str(self.support)
        LOGGER.info("|_ Mathematical support: %s ", math_support)
        num_range = str(self.range)
        LOGGER.info("|_ Numerical range: %s", num_range)

    def compute_samples(
        self,
        n_samples=1,  # type: int
    ):  # noqa: D102
        # type: (...) -> ndarray
        return vstack([marginal.rvs(n_samples) for marginal in self.marginals]).T

    def compute_cdf(
        self,
        vector,  # type: Iterable[float]
    ):  # noqa: D102
        # type: (...) -> ndarray
        return array(
            [self.marginals[index].cdf(value) for index, value in enumerate(vector)]
        )

    def compute_inverse_cdf(
        self,
        vector,  # type: Iterable[float]
    ):  # noqa: D102
        # type: (...) -> ndarray
        return array(
            [self.marginals[index].ppf(value) for index, value in enumerate(vector)]
        )

    @property
    def mean(self):  # noqa: D102
        # type: (...) -> ndarray
        mean = [marginal.mean() for marginal in self.marginals]
        return array(mean)

    @property
    def standard_deviation(self):  # noqa: D102
        # type: (...) -> ndarray
        std = [marginal.std() for marginal in self.marginals]
        return array(std)

    def __create_distributions(
        self,
        distribution,  # type: str
        parameters,  # type: Mapping[str,Union[int,float]]
    ):  # type: (...) -> List[scipy.stats.rv_continuous]
        """Instantiate a SciPy distribution for each random variable component.

        Args:
            distribution: The name of the distribution.
            parameters: The parameters of the distribution.

        Returns:
            The marginal SciPy distributions.
        """
        try:
            sp_dist = getattr(sp_stats, distribution)
        except Exception:
            raise ValueError("%s is an unknown scipy distribution. " % distribution)
        try:
            sp_dist(**parameters)
        except Exception:
            raise ValueError(
                "%s does not take these arguments. "
                "More details on: %s" % (distribution, SP_WEBSITE)
            )
        sp_dist = [sp_dist(**parameters)] * self.dimension
        self.__set_bounds(sp_dist)
        return sp_dist

    def __set_bounds(
        self,
        distributions,  # type: Iterable[scipy.stats.rv_continuous]
        extrema_level=1e-12,  # type: float
    ):  # type: (...) -> None
        """Set mathematical and numerical bounds (= support and range).

        Args:
            distributions: The distributions.
            extrema_level: A quantile level corresponding to the lower
                bound of the numerical random variable range.
        """
        self.math_lower_bound = []
        self.math_upper_bound = []
        self.num_lower_bound = []
        self.num_upper_bound = []
        for distribution in distributions:
            dist_range = distribution.interval(1.0)
            self.math_lower_bound.append(dist_range[0])
            self.math_upper_bound.append(dist_range[1])
            l_b = distribution.ppf(extrema_level)
            u_b = distribution.ppf(1 - extrema_level)
            self.num_lower_bound.append(l_b)
            self.num_upper_bound.append(u_b)
        self.math_lower_bound = array(self.math_lower_bound)
        self.math_upper_bound = array(self.math_upper_bound)
        self.num_lower_bound = array(self.num_lower_bound)
        self.num_upper_bound = array(self.num_upper_bound)

    def _pdf(
        self,
        index,  # type: int
    ):  # noqa: D102
        # type: (...) -> Callable
        def pdf(
            point,  # type: float
        ):
            # type: (...) -> float
            """Probability Density Function (PDF).

            Args:
                point: An evaluation point.

            Returns:
                The PDF value at the evaluation point.
            """
            return self.marginals[index].pdf(point)

        return pdf

    def _cdf(
        self,
        index,  # type: int
    ):  # noqa: D102
        # type: (...) -> Callable
        def cdf(
            level,  # type: float
        ):
            # type: (...) -> float
            """Cumulative Density Function (CDF).

            Args:
                level: A probability level.

            Returns:
                The CDF value for the probability level.
            """
            return self.marginals[index].cdf(level)

        return cdf



# -*- coding: utf-8 -*-
# Copyright (c) 2009-2014, Luke Maurits <luke@maurits.id.au>
# All rights reserved.
# With contributions from:
# * Chris Clark
#  * Klein Stephane
#  * John Filleau
# PTable is forked from original Google Code page in April, 2015, and now
# maintained by Kane Blueriver <kxxoling@gmail.com>.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# Taken from :
# https://github.com/kxxoling/PTable
# https://pypi.python.org/pypi/PTable

"""
A pretty table in the shell or in a file
"""
from __future__ import ( division, 
                        unicode_literals)


from gemseo.third_party.prettytable.prettytable import PrettyTable





# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret, Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Parallel execution of disciplines and functions using multiprocessing
*********************************************************************
"""
from __future__ import division, unicode_literals

import logging
import multiprocessing as mp
import os
import queue
import threading as th
import time
import traceback

LOGGER = logging.getLogger(__name__)


class ParallelExecution(object):
    """Perform a parallel execution of tasks on input values.

    Input values must be a list of independent pointers.
    """

    N_CPUS = mp.cpu_count()

    def __init__(
        self,
        worker_list,
        n_processes=N_CPUS,
        use_threading=False,
        wait_time_between_fork=0,
    ):
        """Constructor.

        :param worker_list: list of objects that perform the tasks
        :param n_processes: maximum number of processors on which to run
        :param use_threading: if True, use Threads instead of processes
            to parallelize the execution
            multiprocessing will copy (serialize) all the disciplines,
            while threading will share all the memory
            This is important to note if you want to execute the same
            discipline multiple times, you shall use multiprocessing
        :param wait_time_between_fork: time waited between two forks of the
            process /Thread
        """
        self.worker_list = worker_list
        self.n_processes = n_processes
        self.use_threading = use_threading

        if not self.use_threading and os.name == "nt":
            raise ValueError(
                "Multiprocessing is not currently supported on"
                " Windows. Please try to use multithreading"
                " instead."
            )

        if use_threading:
            ids = set(id(worker) for worker in worker_list)
            if len(ids) != len(worker_list):
                raise ValueError(
                    "When using multithreading, all workers"
                    " shall be different objects !"
                )
        self.wait_time_between_fork = wait_time_between_fork

    def _run_task_by_index(self, worker_index, input_values):
        """Run a task from an index of discipline.

        Method to run a task from an index of discipline and input local data,
        the purpose is to be used by multiprocessing queues as a task.

        :param worker_index: the index of the worker among self.worker_index
        :param input_values: the input values list
        """
        input_loc = input_values[worker_index]
        worker = self.worker_list[worker_index]
        # return the worker index to order the outputs properly
        output = self._run_task(worker, input_loc)
        return worker_index, output

    def execute(
        self, input_data_list, exec_callback=None, task_submitted_callback=None
    ):
        """Execute all processes.

        :param input_data_list: the input values (list or values)
        :param exec_callback: callback function called with the
            pair (index, outputs) as arguments when an item is retrieved
            from the processing, where index is the associated index
            in input_data_list, of the input used to compute the outputs
        :param task_submitted_callback: callback function called when all the
            tasks are submitted, but not yet done
        """
        n_workers = len(self.worker_list)

        if len(input_data_list) != n_workers:
            raise ValueError(
                "Parallel execution shall be run "
                + "with a list of "
                + "inputs of same size as the list of disciplines"
            )

        if exec_callback is not None and not callable(exec_callback):
            raise TypeError("exec_callback function must be callable !")

        if task_submitted_callback is not None:
            if not callable(task_submitted_callback):
                raise TypeError("task_submitted_callback function must be callable !")

        # Queue for workers
        if self.use_threading:
            queue_in = queue.Queue()
            queue_out = queue.Queue()
        else:
            mananger = mp.Manager()
            queue_in = mananger.Queue()
            queue_out = mananger.Queue()

        def worker():
            """Worker method executes a function while there are args left in the
            queue_in."""
            for args in iter(queue_in.get, None):
                try:
                    function_output = self._run_task_by_index(*args)
                except Exception as err:
                    traceback.print_exc()
                    queue_out.put((args[0], err))
                    queue_in.task_done()
                    continue
                queue_out.put(function_output)
                queue_in.task_done()

        processes = []
        if self.use_threading:
            for _ in range(self.n_processes):
                thread = th.Thread(target=worker)
                thread.daemon = True
                thread.start()
                processes.append(thread)
        else:
            for _ in range(self.n_processes):

                proc = mp.Process(target=worker)
                proc.daemon = True
                proc.start()
                processes.append(proc)

        # fill input queue
        for worker_index in range(n_workers):
            # delay the next processes execution after the first one
            if self.wait_time_between_fork > 0 and worker_index > 0:
                time.sleep(self.wait_time_between_fork)
            queue_in.put((worker_index, input_data_list))

        if task_submitted_callback is not None:
            task_submitted_callback()
        # sort the outputs with the same order as functions
        ordered_outputs = [None] * n_workers
        got_n_outs = 0
        # Retrieve outputs on the fly to call the callbacks, typically
        # iterates progress bar and stores the data in database or cache
        while got_n_outs != n_workers:
            index, output = queue_out.get()
            if isinstance(output, Exception):
                LOGGER.error("Failed to execute task indexed %s", str(index))
                LOGGER.error(output)
            else:
                ordered_outputs[index] = output
                # Call the callback function
                if exec_callback is not None:
                    exec_callback(index, output)
            got_n_outs += 1

        # Tells threads and processes to terminate
        for _ in processes:
            queue_in.put(None)

        # Join processes and threads
        for proc in processes:
            proc.join()

        # Update self.workers objects
        self._update_local_objects(ordered_outputs)

        # Filters outputs, eventually
        filtered_outputs = self._filter_ordered_outputs(ordered_outputs)

        return filtered_outputs

    @staticmethod
    def _filter_ordered_outputs(ordered_outputs):
        """Filters the ordered_outputs.

        Eventually return a subset in the execute method.
        To be overloaded by subclasses.

        :param ordered_outputs: the list of outputs, map of _run_task
           over inputs_list
        """
        return ordered_outputs

    def _update_local_objects(self, ordered_outputs):
        """Update the local objects from parallel results.

        The ordered_outputs contains the stacked outputs of the function
        _run_task() To be overloaded by subclasses.

        :param ordered_outputs: the list of outputs, map of _run_task
            over inputs_list
        """

    @staticmethod
    def _run_task(worker, input_loc):
        """Effectively performs the computation.

        To be overloaded by subclasses.

        :param worker: the worker pointes
        :param input_loc: input of the worker
        """
        if hasattr(worker, "execute"):
            return worker.execute(input_loc)
        if callable(worker):
            return worker(input_loc)
        raise TypeError("cannot handle worker")


class DiscParallelExecution(ParallelExecution):
    """Execute disciplines in parallel."""

    def _update_local_objects(self, ordered_outputs):
        """Update the local objects from parallel results.

        The ordered_outputs contains the stacked outputs of the function
        _run_task()

        :param ordered_outputs: the list of outputs, map of _run_task
            over inputs_list
        """
        for disc, output in zip(self.worker_list, ordered_outputs):
            # Update discipline local data
            disc.local_data = output


class DiscParallelLinearization(ParallelExecution):
    """Linearize disciplines in parallel."""

    def _update_local_objects(self, ordered_outputs):
        """Update the local objects from parallel results.

        The ordered_outputs contains the stacked outputs of the function
        _run_task()

        :param ordered_outputs: the list of outputs, map of _run_task
            over inputs_list
        """
        for disc, output in zip(self.worker_list, ordered_outputs):
            # Update discipline jacobian
            disc.jac = output[1]
            # Update discipline local data in case of execution
            disc.local_data = output[0]

    @staticmethod
    def _run_task(worker, input_loc):
        """Effectively performs the computation.

        To be overloaded by subclasses

        :param worker: the worker pointes
        :param input_loc: input of the worker
        """
        jac = worker.linearize(input_loc)
        return worker.local_data, jac

    @staticmethod
    def _filter_ordered_outputs(ordered_outputs):
        """Filters the ordered_outputs.

        Eventually return a subset in the execute method.
        To be overloaded by subclasses.

        :param ordered_outputs: the list of outputs, map of _run_task
           over inputs_list
        """
        # Only keep the jacobians as outputs, dismiss local_data
        return [out[1] for out in ordered_outputs]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Plot the derivatives of the functions."""
from __future__ import division, unicode_literals

import logging
from typing import Dict, Iterable, Mapping

from matplotlib import pyplot
from matplotlib.figure import Figure
from numpy import arange, atleast_2d, ndarray

from gemseo.post.opt_post_processor import OptPostProcessor

LOGGER = logging.getLogger(__name__)


class GradientSensitivity(OptPostProcessor):
    """Histograms of the derivatives of objective and constraints.

    The plot method considers the derivatives at the last iteration. The iteration can
    be changed, as well as the x and y figure sizes.
    """

    def _plot(
        self,
        iteration=-1,  # type: int
        figsize_x=10,  # type: int
        figsize_y=10,  # type: int
        scale_gradients=False,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            iteration: The iteration to plot the sensitivities;
                if negative, use the optimum.
            figsize_x: The size of figure in the horizontal direction (inches).
            figsize_y: The size of figure in the vertical direction (inches).
            scale_gradients: If True, normalize each gradient
                w.r.t. the design variables.
        """
        if iteration == -1:
            x_ref = self.opt_problem.solution.x_opt
        else:
            x_ref = self.opt_problem.database.get_x_by_iter(iteration)
        grad_dict = self.__get_grad_dict(x_ref, scale_gradients=scale_gradients)

        x_names = self._generate_x_names()

        fig = self.__generate_subplots(
            x_names,
            x_ref,
            grad_dict,
            figsize_x,
            figsize_y,
            scale_gradients=scale_gradients,
        )
        self._add_figure(fig)

    def __get_grad_dict(
        self,
        x_ref,  # type: ndarray
        scale_gradients=False,  # type: bool
    ):  # type: (...) -> Dict[str,ndarray]
        """Create a gradient dictionary from a given iteration.

        Scale it if necessary.

        Args:
            x_ref: The reference value for x.
            scale_gradients: If True, normalize each gradient
                w.r.t. the design variables.

        Returns:
            The gradients of the outputs
            indexed by the names of the output,
            e.g. 'output_name' for a mono-dimensional output,
            or 'output_name_i' for the i-th component of a multi-dimensional output.
        """
        all_funcs = self.opt_problem.get_all_functions_names()
        scale_func = self.opt_problem.design_space.unnormalize_vect
        grad_dict = {}
        for func in all_funcs:
            grad = self.database.get_f_of_x("@{}".format(func), x_ref)
            if grad is not None:
                if len(grad.shape) == 1:
                    if scale_gradients:
                        grad = scale_func(grad, minus_lb=False)
                    grad_dict[func] = grad
                else:
                    n_f, _ = grad.shape
                    for i in range(n_f):
                        if scale_gradients:
                            grad[i, :] = scale_func(grad[i, :], minus_lb=False)
                        grad_dict["{}_{}".format(func, i)] = grad[i, :]
        return grad_dict

    @staticmethod
    def __generate_subplots(
        x_names,  # type: Iterable[str]
        x_ref,  # type: ndarray
        grad_dict,  # type: Mapping[str, ndarray]
        figsize_x=10,  # type: int
        figsize_y=10,  # type: int
        scale_gradients=False,  # type: bool
    ):  # type: (...)-> Figure
        """Generate the gradients subplots from the data.

        Args:
            x_names: The variables names.
            x_ref: The reference value for x.
            grad_dict: The gradients to plot.
            figsize_x: The size of the figure in the horizontal direction (inches).
            figsize_y: The size of the figure in the vertical direction (inches).
            scale_gradients: If True, normalize the gradients w.r.t. the design variables.

        Returns:
            The gradients subplots.

        Raises:
            ValueError: If `grad_dict` is empty.
        """
        n_funcs = len(grad_dict)
        if n_funcs == 0:
            raise ValueError("No gradients to plot at current iteration!")
        nrows = n_funcs // 2
        if 2 * nrows < n_funcs:
            nrows += 1
        ncols = 2
        fig, axes = pyplot.subplots(
            nrows=nrows,
            ncols=2,
            sharex=True,
            sharey=False,
            figsize=(figsize_x, figsize_y),
        )
        i = 0
        j = -1

        axes = atleast_2d(axes)
        n_subplots = len(axes) * len(axes[0])
        abscissa = arange(len(x_ref))
        x_labels = [str(x_id) for x_id in x_names]
        for func, grad in sorted(grad_dict.items()):
            j += 1
            if j == ncols:
                j = 0
                i += 1
            axe = axes[i][j]
            axe.bar(abscissa, grad, color="blue", align="center")
            axe.set_title(func)
            axe.set_xticklabels(x_labels, fontsize=12, rotation=90)
            axe.set_xticks(abscissa)
            # Update y labels spacing
            vis_labels = [
                label for label in axe.get_yticklabels() if label.get_visible() is True
            ]
            pyplot.setp(vis_labels[::2], visible=False)

        if len(grad_dict) < n_subplots:
            # xlabel must be written with the same fontsize on the 2 columns
            j += 1
            #             if j == ncols: Seems impossible to reach
            #                 j = 0
            #                 i += 1
            axe = axes[i][j]
            axe.set_xticklabels(x_labels, fontsize=12, rotation=90)
            axe.set_xticks(abscissa)

        if scale_gradients:
            fig.suptitle(
                "Derivatives of objective and constraints"
                + " with respect to design variables.\n \nNormalized Design Space.",
                fontsize=14,
            )
        else:
            fig.suptitle(
                "Derivatives of objective and constraints"
                + " with respect to design variables",
                fontsize=14,
            )
        return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""This module contains the base class for classification algorithms.

The :mod:`~gemseo.mlearning.classification.classification` module
implements classification algorithms,
whose goal is to assess the membership of input data to classes.

Classification algorithms provide methods for predicting classes of new input data,
as well as predicting the probabilities of belonging to each of the classes
wherever possible.

This concept is implemented through the :class:`.MLClassificationAlgo` class
which inherits from the :class:`.MLSupervisedAlgo` class.
"""
from __future__ import division, unicode_literals

from typing import Dict, Iterable, List, Optional, Sequence, Union

from numpy import ndarray, unique, zeros

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import DataType, MLAlgoParameterType, TransformerType
from gemseo.mlearning.core.supervised import MLSupervisedAlgo
from gemseo.mlearning.core.supervised import (
    SavedObjectType as MLSupervisedAlgoSavedObjectType,
)

SavedObjectType = Union[
    MLSupervisedAlgoSavedObjectType, Sequence[str], Dict[str, ndarray], int
]


class MLClassificationAlgo(MLSupervisedAlgo):
    """Classification Algorithm.

    Inheriting classes shall implement the :meth:`!MLSupervisedAlgo._fit` and
    :meth:`!MLClassificationAlgo._predict` methods, and
    :meth:`!MLClassificationAlgo._predict_proba_soft` method if possible.

    Attributes:
        n_classes (int): The number of classes.
    """

    def __init__(
        self,
        data,  # type: Dataset
        transformer=MLSupervisedAlgo.DEFAULT_TRANSFORMER,  # type: TransformerType
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        **parameters  # type: MLAlgoParameterType
    ):  # type: (...) -> None
        super(MLClassificationAlgo, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            **parameters
        )
        self.n_classes = None

    def learn(
        self,
        samples=None,  # type: Optional[List[int]]
    ):  # type: (...) -> None
        output_data = self.learning_set.get_data_by_names(self.output_names, False)
        self.n_classes = unique(output_data).shape[0]
        super(MLClassificationAlgo, self).learn(samples)

    @MLSupervisedAlgo.DataFormatters.format_input_output
    def predict_proba(
        self,
        input_data,  # type: DataType
        hard=True,  # type: bool
    ):  # type: (...)-> ndarray
        """Predict the probability of belonging to each cluster from input data.

        The user can specified these input data either as a numpy array,
        e.g. :code:`array([1., 2., 3.])`
        or as a dictionary,
        e.g.  :code:`{'a': array([1.]), 'b': array([2., 3.])}`.

        If the numpy arrays are of dimension 2,
        their i-th rows represent the input data of the i-th sample;
        while if the numpy arrays are of dimension 1,
        there is a single sample.

        The type of the output data and the dimension of the output arrays
        will be consistent
        with the type of the input data and the size of the input arrays.

        Args:
            input_data: The input data.
            hard: Whether clustering should be hard (True) or soft (False).

        Returns:
            The probability of belonging to each cluster.
        """
        return self._predict_proba(input_data, hard)

    def _predict_proba(
        self,
        input_data,  # type: ndarray
        hard=True,  # type: bool
    ):  # type: (...)-> ndarray
        """Predict the probability of belonging to each class.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).
            hard: Whether clustering should be hard (True) or soft (False).

        Returns:
            The probability of belonging to each class
                with shape (n_samples, n_classes).
        """
        if hard:
            probas = self._predict_proba_hard(input_data)
        else:
            probas = self._predict_proba_soft(input_data)
        return probas

    def _predict_proba_hard(
        self,
        input_data,  # type: ndarray
    ):  # type: (...)-> ndarray
        """Return 1 if the data belongs to a class, 0 otherwise.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            The indicator of belonging to each class with shape (n_samples, n_classes).
        """
        n_samples = input_data.shape[0]
        prediction = self._predict(input_data).astype(int)
        n_outputs = prediction.shape[1]
        probas = zeros((n_samples, self.n_classes, n_outputs))
        for n_sample in range(prediction.shape[0]):
            for n_output in range(n_outputs):
                probas[n_sample, prediction[n_sample, n_output], n_output] = 1
        return probas

    def _predict_proba_soft(
        self,
        input_data,  # type: ndarray
    ):  # type: (...)-> ndarray
        """Predict the probability of belonging to each class.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            The probability of belonging to each class
                with shape (n_samples, n_classes).
        """
        raise NotImplementedError

    def _get_objects_to_save(self):  # type: (...) -> SavedObjectType
        objects = super(MLClassificationAlgo, self)._get_objects_to_save()
        objects["n_classes"] = self.n_classes
        return objects



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret
#                 Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""The disciplines for the MDO problem proposed by Sellar et al. in

Sellar, R., Batill, S., & Renaud, J. (1996).
Response surface based, concurrent subspace optimization
for multidisciplinary system design.
In 34th aerospace sciences meeting and exhibit (p. 714).

The MDO problem is written as follows:

.. math::

   \begin{aligned}
   \text{minimize the objective function }&obj=x_{local}^2 + x_{shared,2}
   +y_1^2+e^{-y_2} \\
   \text{with respect to the design variables }&x_{shared},\,x_{local} \\
   \text{subject to the general constraints }
   & c_1 \leq 0\\
   & c_2 \leq 0\\
   \text{subject to the bound constraints }
   & -10 \leq x_{shared,1} \leq 10\\
   & 0 \leq x_{shared,2} \leq 10\\
   & 0 \leq x_{local} \leq 10.
   \end{aligned}

where the coupling variables are

.. math::

    \text{Discipline 1: } y_1 = \sqrt{x_{shared,1}^2 + x_{shared,2} +
     x_{local} - 0.2\,y_2},

and

.. math::

    \text{Discipline 2: }y_2 = |y_1| + x_{shared,1} + x_{shared,2}.

and where the general constraints are

.. math::

   c_1 = 3.16 - y_1^2

   c_2 = y_2 - 24

This module implements three disciplines to compute the different coupling variables,
constraints and objective:

- :class:`.Sellar1`:
  this :class:`.MDODiscipline` computes :math:`y_1`
  from :math:`y_2`, :math:`x_{shared,1}`, :math:`x_{shared,2}` and :math:`x_{local}`.
- :class:`.Sellar2`:
  this :class:`.MDODiscipline` computes :math:`y_2`
  from :math:`y_1`, :math:`x_{shared,1}` and :math:`x_{shared,2}`.
- :class:`.SellarSystem`:
  this :class:`.MDODiscipline` computes both objective and constraints
  from :math:`y_1`, :math:`y_2`, :math:`x_{local}` and :math:`x_{shared,2}`.
"""
from __future__ import division, unicode_literals

from cmath import exp, sqrt
from typing import Dict, Iterable, List, Optional

from numpy import array, atleast_2d, complex128, ndarray, ones, zeros

from gemseo.core.discipline import MDODiscipline

Y_1 = "y_1"
Y_2 = "y_2"
X_SHARED = "x_shared"
X_LOCAL = "x_local"
OBJ = "obj"
C_1 = "c_1"
C_2 = "c_2"
R_1 = "r_1"
R_2 = "r_2"


def get_inputs(
    names=None,  # type: Optional[Iterable[str]]
):  # type: (...) -> Dict[str,ndarray]
    """Generate an initial solution for the MDO problem.

    Args:
        names: The names of the discipline inputs.

    Returns:
        The default values of the discipline inputs.
    """
    inputs = {
        X_LOCAL: array([0.0], dtype=complex128),
        X_SHARED: array([1.0, 0.0], dtype=complex128),
        Y_1: ones(1, dtype=complex128),
        Y_2: ones(1, dtype=complex128),
    }
    if names is None:
        return inputs
    return {name: inputs[name] for name in names}


class SellarSystem(MDODiscipline):
    """The discipline to compute the objective and constraints of the Sellar problem."""

    def __init__(self):  # type: (...) -> None
        super(SellarSystem, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = get_inputs()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):  # type: (...) -> None
        x_local, x_shared, y_1, y_2 = self.get_inputs_by_name(
            [X_LOCAL, X_SHARED, Y_1, Y_2]
        )
        obj = array([self.compute_obj(x_local, x_shared, y_1, y_2)], dtype=complex128)
        c_1 = array([self.compute_c_1(y_1)], dtype=complex128)
        c_2 = array([self.compute_c_2(y_2)], dtype=complex128)
        self.store_local_data(obj=obj, c_1=c_1, c_2=c_2)

    @staticmethod
    def compute_obj(
        x_local,  # type: ndarray
        x_shared,  # type: ndarray
        y_1,  # type: ndarray
        y_2,  # type: ndarray
    ):  # type: (...) -> float
        """Evaluate the objective :math:`obj`.

        Args:
            x_local: The design variables local to the first discipline.
            x_shared: The shared design variables.
            y_1: The coupling variable coming from the first discipline.
            y_2: The coupling variable coming from the second discipline.

        Returns:
            The value of the objective :math:`obj`.
        """
        return x_local[0] ** 2 + x_shared[1] + y_1[0] ** 2 + exp(-y_2[0])

    @staticmethod
    def compute_c_1(
        y_1,  # type: ndarray
    ):  # type: (...) -> float
        """Evaluate the constraint :math:`c_1`.

        Args:
            y_1: The coupling variable coming from the first discipline.

        Returns:
            The value of the constraint :math:`c_1`.
        """
        return 3.16 - y_1[0] ** 2

    @staticmethod
    def compute_c_2(
        y_2,  # type: ndarray
    ):  # type: (...) -> float
        """Evaluate the constraint :math:`c_2`.

        Args:
            y_2: The coupling variable coming from the second discipline.

        Returns:
            The value of the constraint :math:`c_2`.
        """
        return y_2[0] - 24.0

    def _compute_jacobian(
        self,
        inputs=None,  # type: Optional[Iterable[str]]
        outputs=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> None
        self._init_jacobian(inputs, outputs, with_zeros=True)
        x_local, _, y_1, y_2 = self.get_inputs_by_name([X_LOCAL, X_SHARED, Y_1, Y_2])
        self.jac[C_1][Y_1] = atleast_2d(array([-2.0 * y_1]))
        self.jac[C_2][Y_2] = ones((1, 1))
        self.jac[OBJ][X_LOCAL] = atleast_2d(array([2.0 * x_local[0]]))
        self.jac[OBJ][X_SHARED] = atleast_2d(array([0.0, 1.0]))
        self.jac[OBJ][Y_1] = atleast_2d(array([2.0 * y_1[0]]))
        self.jac[OBJ][Y_2] = atleast_2d(array([-exp(-y_2[0])]))


class Sellar1(MDODiscipline):
    """The discipline to compute the coupling variable :math:`y_1`."""

    def __init__(
        self,
        residual_form=False,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            residual_form: If True,
                then only residuals are computed,
                not the coupling variables.
        """
        self.residual_form = residual_form
        super(Sellar1, self).__init__(auto_detect_grammar_files=True)
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY
        if residual_form:
            self.residual_variables = [R_1]
            self.output_grammar.remove_item(Y_1)

        else:
            self.output_grammar.remove_item(R_1)
            self.input_grammar.remove_item(Y_1)

        self.default_inputs = get_inputs(self.input_grammar.get_data_names())

    def get_attributes_to_serialize(self):  # type: (...) -> List[str]
        base_d = super(Sellar1, self).get_attributes_to_serialize()
        base_d.append("residual_form")
        return base_d

    def _run(self):  # type: (...) -> None
        x_local, x_shared, y_2 = self.get_inputs_by_name([X_LOCAL, X_SHARED, Y_2])

        if self.residual_form:
            y_1 = self.get_inputs_by_name(Y_1)
            r_1_out = array(
                [self.compute_r_1(x_local, x_shared, y_1, y_2)], dtype=complex128
            )
            self.store_local_data(r_1=r_1_out)
        else:
            # functional form
            y_1_out = array(
                [self.compute_y_1(x_local, x_shared, y_2)], dtype=complex128
            )
            self.store_local_data(y_1=y_1_out)

    @staticmethod
    def compute_y_1(
        x_local,  # type: ndarray
        x_shared,  # type: ndarray
        y_2,  # type: ndarray
    ):  # type: (...) -> float
        """Evaluate the first coupling equation in functional form.

        Args:
            x_local: The design variables local to first discipline.
            x_shared: The shared design variables.
            y_2: The coupling variable coming from the second discipline.

        Returns:
            The value of the coupling variable :math:`y_1`.
        """
        return sqrt(x_shared[0] ** 2 + x_shared[1] + x_local[0] - 0.2 * y_2[0])

    @staticmethod
    def compute_r_1(
        x_local,  # type: ndarray
        x_shared,  # type: ndarray
        y_1,  # type: ndarray
        y_2,  # type: ndarray
    ):  # type: (...) -> float
        """Evaluate the first coupling equation in residual form.

        Args:
            x_local: The design variables local to the first discipline.
            x_shared: The shared design variables.
            y_1: The coupling variable coming from the first discipline.
            y_2: The coupling variable coming from the second discipline.

        Returns:
            The value of the residues related to the first discipline.
        """
        return Sellar1.compute_y_1(x_local, x_shared, y_2) - y_1[0]

    def _compute_jacobian(
        self,
        inputs=None,  # type: Optional[Iterable[str]]
        outputs=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> None
        self._init_jacobian(inputs, outputs, with_zeros=True)
        x_local, x_shared, y_2 = self.get_inputs_by_name([X_LOCAL, X_SHARED, Y_2])

        inv_denom = 1.0 / (self.compute_y_1(x_local, x_shared, y_2))
        self.jac[Y_1] = {}
        self.jac[Y_1][X_LOCAL] = atleast_2d(array([0.5 * inv_denom]))
        self.jac[Y_1][X_SHARED] = atleast_2d(
            array([x_shared[0] * inv_denom, 0.5 * inv_denom])
        )
        self.jac[Y_1][Y_2] = atleast_2d(array([-0.1 * inv_denom]))

        if self.residual_form:
            self.jac[R_1][X_LOCAL] = atleast_2d(self.jac[Y_1][X_LOCAL])
            self.jac[R_1][X_SHARED] = atleast_2d(self.jac[Y_1][X_SHARED])
            self.jac[R_1][Y_2] = atleast_2d(self.jac[Y_1][Y_2])
            self.jac[R_1][Y_1] = -ones((1, 1))
            del self.jac[Y_1]


class Sellar2(MDODiscipline):
    """The discipline to compute the coupling variable :math:`y_2`."""

    def __init__(
        self,
        residual_form=False,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            residual_form: If True,
                then only residuals are computed,
                not the coupling variables.
        """
        self.residual_form = residual_form
        super(Sellar2, self).__init__(auto_detect_grammar_files=True)
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY
        if residual_form:
            self.residual_variables = [R_2]
            self.output_grammar.remove_item(Y_2)
        else:
            self.output_grammar.remove_item(R_2)
            self.input_grammar.remove_item(Y_2)
        self.default_inputs = get_inputs(self.input_grammar.get_data_names())

    def get_attributes_to_serialize(self):
        base_d = super(Sellar2, self).get_attributes_to_serialize()
        base_d.append("residual_form")
        return base_d

    def _run(self):  # type: (...) -> None
        x_shared, y_1 = self.get_inputs_by_name([X_SHARED, Y_1])

        if self.residual_form:
            y_2 = self.get_inputs_by_name(Y_2)
            r_2_out = array([self.compute_r_2(x_shared, y_1, y_2)], dtype=complex128)
            self.store_local_data(r_2=r_2_out)
        else:
            # functional form
            y_2_out = array([self.compute_y_2(x_shared, y_1)], dtype=complex128)
            self.store_local_data(y_2=y_2_out)

    def _compute_jacobian(
        self,
        inputs=None,  # type: Optional[Iterable[str]]
        outputs=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> None
        self._init_jacobian(inputs, outputs, with_zeros=True)
        y_1 = self.get_inputs_by_name(Y_1)
        self.jac[Y_2] = {}
        self.jac[Y_2][X_LOCAL] = zeros((1, 1))
        self.jac[Y_2][X_SHARED] = ones((1, 2))
        if y_1[0] < 0.0:
            self.jac[Y_2][Y_1] = -ones((1, 1))
        elif y_1[0] == 0.0:
            self.jac[Y_2][Y_1] = zeros((1, 1))
        else:
            self.jac[Y_2][Y_1] = ones((1, 1))

        if self.residual_form:
            self.jac[R_2][X_LOCAL] = self.jac[Y_2][X_LOCAL]
            self.jac[R_2][X_SHARED] = self.jac[Y_2][X_SHARED]
            self.jac[R_2][Y_1] = self.jac[Y_2][Y_1]
            self.jac[R_2][Y_2] = -ones((1, 1))
            del self.jac[Y_2]

    @staticmethod
    def compute_y_2(
        x_shared,  # type: ndarray
        y_1,  # type: ndarray
    ):  # type: (...) -> float
        """Evaluate the second coupling equation in functional form.

        Args:
            x_shared: The shared design variables.
            y_1: The coupling variable coming from the first discipline.

        Returns:
            The value of the coupling variable :math:`y_2`.
        """
        out = x_shared[0] + x_shared[1]
        if y_1[0].real == 0:
            y_2 = out
        elif y_1[0].real > 0:
            y_2 = y_1[0] + out
        else:
            y_2 = -y_1[0] + out
        return y_2

    @staticmethod
    def compute_r_2(
        x_shared,  # type: ndarray
        y_1,  # type: ndarray
        y_2,  # type: ndarray
    ):  # type: (...) -> float
        """Evaluate the second coupling equation in residual form.

        Args:
            x_shared: The shared design variables.
            y_1: The coupling variable coming from the first discipline.
            y_2: The coupling variable coming from the second discipline.

        Returns:
            The value of the residues related to the second discipline.
        """
        return Sellar2.compute_y_2(x_shared, y_1) - y_2[0]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Self Organizing Maps to display high dimensional design spaces."""

from __future__ import division, unicode_literals

import logging
from math import floor, sqrt

import matplotlib
from numpy import array, bincount, float64, int32, isnan, logical_not
from numpy import max as np_max
from numpy import mean, mgrid
from numpy import min as np_min
from numpy import ndarray, nonzero, unique, where, zeros
from pylab import plt

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.post.core.colormaps import PARULA
from gemseo.post.opt_post_processor import OptPostProcessor
from gemseo.third_party import sompy

LOGGER = logging.getLogger(__name__)


class SOM(OptPostProcessor):
    """Self organizing map clustering optimization history.

    Options of the plot method are the figure width and height, and the x- and y-
    numbers of cells in the SOM.
    """

    def __init__(
        self,
        opt_problem,  # type: OptimizationProblem
    ):  # type: (...) -> None
        super(SOM, self).__init__(opt_problem)
        self.som = None
        self.cmap = PARULA

    @staticmethod
    def __build_som_from_vars(
        x_vars,  # type: ndarray
        som_grid_nx=5,  # type:int
        som_grid_ny=5,  # type:int
        initmethod="pca",  # type:str
        verbose="off",  # type:str
    ):  # type: (...) -> SOM
        """Builds the SOM from the design variables history.

        Args:
            x_vars: The design variables history (n_iter,n_dv).
            som_grid_nx: The number of neurons in the x direction.
            som_grid_ny: The number of neurons in the y direction.
            initmethod: The initialization method for the SOM.
            verbose: The verbose for SOM training.

        Returns:
            The self organizing map
        """

        LOGGER.info("Building Self Organizing Map from optimization history:")
        LOGGER.info("    Number of neurons in x direction = %s", str(som_grid_nx))
        LOGGER.info("    Number of neurons in y direction = %s", str(som_grid_ny))
        var_som = sompy.SOM(
            "som",
            x_vars,
            mapsize=[som_grid_ny + 1, som_grid_nx + 1],
            norm_method="var",
            initmethod=initmethod,
        )
        var_som.init_map()
        var_som.train(n_job=1, shared_memory="no", verbose=verbose)
        return var_som

    def _plot(
        self,
        n_x=4,  # type: int
        n_y=4,  # type: int
        width=12,  # type: int
        height=18,  # type: int
        annotate=False,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            n_x: The number of grids in x.
            n_y: The number of grids in y.
            width: The width of the figure (in inches).
            height: The height of the figure (in inches).
            annotate: If True, add label of neuron value to SOM plot.
        """
        criteria_list = [
            self.opt_problem.get_objective_name()
        ] + self.opt_problem.get_constraints_names()
        all_data = self.database.get_all_data_names()
        # Ensure that the data is available in the database
        for crit in criteria_list:
            if crit not in all_data:
                criteria_list.remove(crit)
        figure = plt.figure(figsize=(width, height), dpi=80)
        figure.suptitle("Self Organizing Maps of the design space", fontsize=14)
        subplot_number = 0
        self.__compute(n_x, n_y)
        for criteria in criteria_list:
            f_hist, _ = self.database.get_complete_history(
                ["SOM_i", "SOM_j", "SOM_indx", criteria]
            )
            if isinstance(f_hist[0][3], ndarray):
                dim_val = f_hist[0][3].size
                for _ in range(dim_val):
                    subplot_number += 1

            else:
                subplot_number += 1

        grid_size_x = 3
        grid_size_y = subplot_number // grid_size_x
        if (subplot_number % grid_size_x) > 0:
            grid_size_y += 1

        fig_indx = 1
        for criteria in criteria_list:
            f_hist, _ = self.database.get_complete_history(
                ["SOM_i", "SOM_j", "SOM_indx", criteria]
            )
            if isinstance(f_hist[0][3], ndarray):
                dim_val = f_hist[0][3].size
                for k in range(dim_val):
                    f_hist_scalar = []
                    for f_h in f_hist:
                        scal_list = f_h[0:3]
                        scal_list.append(f_h[3][k])
                        f_hist_scalar.append(scal_list)
                    criteria_name = criteria + "_" + str(k)
                    self.__plot_som_from_scalar_data(
                        f_hist_scalar,
                        criteria_name,
                        fig_indx,
                        grid_size_x=grid_size_x,
                        grid_size_y=grid_size_y,
                        annotate=annotate,
                    )
                    fig_indx += 1

            else:
                self.__plot_som_from_scalar_data(
                    f_hist,
                    criteria,
                    fig_indx,
                    grid_size_x=grid_size_x,
                    grid_size_y=grid_size_y,
                    annotate=annotate,
                )
                fig_indx += 1

        self._add_figure(figure)

    def __plot_som_from_scalar_data(
        self,
        f_hist_scalar,  # type: ndarray
        criteria,  # type: str
        fig_indx,  # type: int
        grid_size_x=3,  # type: int
        grid_size_y=20,  # type: int
        annotate=False,  # type: bool
    ):
        """Builds the SOM plot after computation for a given criteria.

        Args:
            criteria: The criteria to show.
            f_hist_scalar: The scalar data to show.
            fig_indx: The axe index in the figure.
            grid_size_x: The number of SOMs in the grid on the x axis.
            grid_size_y: The number of SOMs in the grid on the y axis.
            annotate: If True, add label with average value of neural.
        """
        f_hist = array(f_hist_scalar).T.real
        unique_ind = unique(f_hist[2, :])
        average = {}
        for _, som_id in enumerate(unique_ind):
            where_somid = where(f_hist[2, :] == som_id)[0]
            ranges_of_uniques = f_hist[3, where_somid]
            average[som_id] = mean(ranges_of_uniques)

        ijshape = array((np_max(f_hist[0, :]), np_max(f_hist[1, :])), dtype=int32)
        mat_ij = zeros(ijshape, dtype=float64)
        mat_ij[:, :] = float("nan")
        for itr in range(f_hist.shape[-1]):
            i, j, somindx, _ = f_hist[:, itr]
            mat_ij[int(i) - 1, int(j) - 1] = average[somindx]
        empty = isnan(mat_ij)
        non_empty = logical_not(empty)
        axe = plt.subplot(grid_size_y, grid_size_x, fig_indx)
        minv = np_min(mat_ij[non_empty])
        maxv = np_max(mat_ij[non_empty])
        self.out_data_dict[fig_indx] = mat_ij
        im1 = axe.imshow(
            mat_ij,
            vmin=minv - 0.01 * abs(minv),
            vmax=maxv + 0.01 * abs(maxv),
            cmap=self.cmap,
            interpolation="nearest",
            aspect="auto",
        )  # "spectral" "hot" "RdBu_r"

        if annotate:
            crit_format = "%1.2g"
            for i in range(mat_ij.shape[0]):
                for j in range(mat_ij.shape[0]):
                    _ = axe.text(
                        j,
                        i,
                        crit_format % mat_ij[i, j],
                        ha="center",
                        va="center",
                        color="w",
                        fontsize=7,
                    )

        axe.set_title(criteria, fontsize=12)
        cax, kwa = matplotlib.colorbar.make_axes([axe])
        plt.colorbar(im1, cax=cax, **kwa)
        im1.axes.get_xaxis().set_visible(False)
        im1.axes.get_yaxis().set_visible(False)
        return axe

    def __compute(
        self,
        som_grid_nx=5,  # type: int
        som_grid_ny=5,  # type: int
    ):
        """Build the SOM from optimization history.

        Args:
            som_grid_nx: The number of neurons in the x direction.
            som_grid_ny: The number of neurons in the y direction.
        """
        x_history = self.database.get_x_history()
        x_vars = array(x_history).real
        self.som = self.__build_som_from_vars(x_vars, som_grid_nx, som_grid_ny)
        som_cluster_index = self.som.project_data(x_vars)
        som_coord = array(self.som.ind_to_xy(som_cluster_index), dtype=int32)
        coord_2d_offset = self.__coord2d_to_coords_offsets(som_coord)
        self.out_data_dict["SOM"] = coord_2d_offset
        for i, x_vars in enumerate(x_history):
            self.database.store(
                x_vars,
                {
                    "SOM_indx": som_cluster_index[i],
                    "SOM_i": som_coord[i, 0],
                    "SOM_j": som_coord[i, 1],
                    "SOM_x": coord_2d_offset[i, 0],
                    "SOM_y": coord_2d_offset[i, 1],
                },
            )

    @staticmethod
    def __coord2d_to_coords_offsets(
        som_coord,  # type: ndarray
        max_ofset=0.6,  # type: float
    ):  # type: (...) -> ndarray
        """Takes a coord array from SOM and adds an offset to the coordinates of the
        elements in the cluster so that they can be distinguished at display.

        Args:
            som_coord: The SOM coordinates.
            max_ofset: The maximum offset of the grid.

        Returns:
            The coordinates.
        """
        coord_2d = som_coord[:, :2]
        coord_2d_offset = array(coord_2d, dtype=float64)
        coord_indx = som_coord[:, -1]
        y_vars = bincount(coord_indx)
        i = nonzero(y_vars)[0]
        uniques_occ = array(list(zip(i, y_vars[i])))
        unique_indx = uniques_occ[:, 0]
        max_occ = np_max(uniques_occ[:, 1])
        max_subarr_size = floor(sqrt(max_occ)) + 1
        dxdy_max = max_ofset / (max_subarr_size - 1)
        for grp in unique_indx:
            inds_of_grp = where(coord_indx == grp)[0]
            subarr_size = sqrt(len(inds_of_grp))
            if floor(subarr_size) < subarr_size:
                subarr_size = floor(subarr_size) + 1
            else:
                subarr_size = floor(subarr_size)
            # Otherwise single individual then no need to build a grid
            if subarr_size > 1:
                grid = mgrid[0:subarr_size, 0:subarr_size] * dxdy_max
                gridx = grid[0, :, :].flatten()
                gridy = grid[1, :, :].flatten()
                for k, ind_in_grp in enumerate(inds_of_grp):
                    coord_2d_offset[ind_in_grp, 0] += gridx[k]
                    coord_2d_offset[ind_in_grp, 1] += gridy[k]
        return coord_2d_offset
